#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

# ruff: noqa: D100   # auto-generated file (Missing docstring in public module)
# ruff: noqa: D101   # auto-generated file (Missing docstring in public class)
# ruff: noqa: E501   # auto-generated file (Line too long)


from __future__ import annotations

import json
import requests
import urllib
import urllib3
from ibm_watsonx_data_integration.common.json_patch_format import prepare_json_patch_payload
from ibm_watsonx_data_integration.common.models import BaseModel, CollectionModel, CollectionModelResults
from ibm_watsonx_data_integration.common.utils import get_params_from_swagger
from pathlib import Path
from pydantic import ConfigDict, Field
from typing import TYPE_CHECKING, Any, ClassVar
from typing_extensions import override

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.cpd_models import Project
    from ibm_watsonx_data_integration.platform import Platform


class ConnectionsServiceInfo(BaseModel):
    configuration: dict[str, Any] | None = Field(None, repr=False)
    failure_message: str | None = Field(
        None, description="A message indicating the cause if the service is not running correctly", repr=False
    )
    service_name: str | None = Field(None, description="The name of the service", repr=True)
    status: str | None = Field(
        None, description="An overall status indicating whether the service is running correctly", repr=False
    )
    timestamp: str | None = Field(
        None,
        description="The timestamp when the information was retrieved (in format YYYY-MM-DDTHH:mm:ssZ or YYYY-MM-DDTHH:mm:ss.sssZ, matching the date-time format as specified by RFC 3339)",
        repr=False,
    )
    version: str | None = Field(None, description="The service version string", repr=True)

    model_config = ConfigDict(frozen=True)


class FunctionalID(BaseModel):
    description: str | None = Field(None, description="Description of this functional.", repr=False)
    entitlements: list[str] | None = Field(None, description="The list of names of relevant entitlements.", repr=False)
    functional_id: str | None = Field(None, description="the ID of function.", repr=True)
    label: str | None = Field(None, description="the label of this functional.", repr=False)
    tools: list[str] | None = Field(None, description="Allowed tools that consume this functional ID.", repr=False)

    model_config = ConfigDict(frozen=True)


class DirectoryAsset(BaseModel):
    path: str | None = Field(None, description="Folder ID that the connection asset lives in.", repr=False)

    model_config = ConfigDict(frozen=True)


class DatasourceTypeProperty(BaseModel):
    default_value: str | None = Field(
        None, description="The default value for the property if the value is not otherwise specified.", repr=False
    )
    description: str | None = Field(None, description="The description for the property.", repr=False)
    group: str | None = Field(None, description="A classification group for the property.", repr=False)
    hidden: bool | None = Field(
        None, description="Whether the property should be displayed in a user interface.", repr=False
    )
    label: str | None = Field(None, description="The label for the property.", repr=False)
    masked: bool | None = Field(
        None,
        description="Whether the property should be masked. For example, when the property is a password.",
        repr=False,
    )
    name: str | None = Field(None, description="The property name.", repr=True)
    required: bool | None = Field(None, description="Whether the property is required.", repr=False)
    type: str | None = Field(None, description="The type of the property.", repr=False)
    values: list[dict] | None = Field(
        None,
        description="If the property type is enum, the list of enumerated values that the property can take.",
        repr=False,
    )

    model_config = ConfigDict(frozen=True)


class ConnectionInteractionProperties(BaseModel):
    source: list[DatasourceTypeProperty] | None = Field(
        None, description="The properties that can be set for a source interaction.", repr=False
    )
    target: list[DatasourceTypeProperty] | None = Field(
        None, description="The properties that can be set for a target interaction.", repr=False
    )

    model_config = ConfigDict(frozen=True)


class ConnectionMetadata(BaseModel):
    asset_category: str | None = Field(None, description='The asset category ("USER" or "SYSTEM")', repr=False)
    asset_id: str | None = Field(None, description="The ID of the asset", repr=False)
    asset_type: str | None = Field(None, description="The type of the asset", repr=False)
    catalog_id: str | None = Field(
        None,
        description="The ID of the catalog which contains the asset. catalog_id, project_id or spaceid is required.",
        repr=False,
    )
    create_time: str | None = Field(
        None,
        description="The timestamp when the asset was created (in format YYYY-MM-DDTHH:mm:ssZ or YYYY-MM-DDTHH:mm:ss.sssZ, matching the date-time format as specified by RFC 3339)",
        repr=False,
    )
    creator_id: str | None = Field(None, description="The IAM ID of the user that created the asset", repr=False)
    href: str | None = Field(None, description="URL that can be used to get the asset.", repr=False)
    owner_id: str | None = Field(None, description="The IAM ID of the user that owns the asset", repr=False)
    project_id: str | None = Field(
        None,
        description="The ID of the project which contains the asset. catalog_id, project_id or spaceid is required.",
        repr=False,
    )
    resource_key: str | None = Field(
        None, description="Optional external unique key for assets that supports it", repr=False
    )
    space_id: str | None = Field(
        None,
        description="The ID of the space which contains the asset. catalog_id, project_id or spaceid is required.",
        repr=False,
    )
    tags: list[str] | None = Field(None, repr=False)


class Connection(BaseModel):
    EXPOSED_DATA_PATH: ClassVar[dict] = {"entity": {}}
    asset_category: str | None = Field(None, description="The asset category", repr=False, frozen=True)
    data_source_definition_asset_id: str | None = Field(
        None,
        description='The id of the data source definition asset related to connection. For example "cfdcb449-1204-44ba-baa6-9a8a878e6aa7".',
        repr=False,
        frozen=True,
    )
    data_source_definition_asset_name: str | None = Field(
        None,
        description='The name of the data source definition asset related to connection. For example "Data privacy profile for DB2".',
        repr=False,
    )
    datasource_type: str = Field(
        description='The id or the name of the data source type to connect to. For example "cfdcb449-1204-44ba-baa6-9a8a878e6aa7" or "db2".',
        repr=False,
    )
    description: str | None = Field(None, description="The description of the connection.", repr=False)
    gateway_id: str | None = Field(
        None,
        description='The id of the secure gateway to use with the connection. A Secure Gateway is needed when connecting to an on-premises data source. This is the id of the Secure Gateway created with the SecureGateway Service. Your Secure Gateway Client running on-premises must be connected to the gateway with this Id. For example, "E9oXGRIhv1e_prod_ng".',
        repr=False,
    )
    interaction_properties: ConnectionInteractionProperties | None = Field(None, repr=False)
    metadata: ConnectionMetadata | None = Field(None, repr=False)
    name: str = Field(description="The name of the connection.", repr=True)
    owner_id: str | None = Field(
        None,
        description="Owner or creator of connection.  Provided when a service ID token is used to create connection.",
        repr=False,
        frozen=True,
    )
    properties: dict | None = Field(None, repr=False)
    ref_asset_id: str | None = Field(
        None,
        description="The ID of the connection in reference catalog that this connection refers to for properties values.",
        repr=False,
    )
    ref_catalog_id: str | None = Field(
        None, description="The ID of the catalog that this connection refers to for properties values.", repr=False
    )
    resource_key: str | None = Field(
        None,
        description="Resource key that should be set in connection asset metadata record instead of using a calculated one by the service.",
        repr=False,
    )
    source_system: dict[str, dict[str, Any]] | None = Field(None, repr=False)
    source_system_history: list[dict[str, dict[str, Any]]] | None = Field(None, repr=False)
    tags: list[str] | None = Field(None, repr=False)

    def __init__(self, platform: Platform = None, project: Project = None, **connection_json: dict) -> None:
        """The __init__ of the Connection class.

        Args:
            connection_json: The JSON for the Connection.
            platform: The Platform object.
            project: The Project object.
        """
        super().__init__(**connection_json)
        self._platform = platform
        self._project = project
        self._origin = self.model_dump()

    @staticmethod
    def _create(
        project: Project,
        name: str,
        datasource_type: DatasourceType,
        description: str | None = None,
        properties: dict | None = None,
        test: bool = True,
    ) -> Connection:
        data = {
            "name": name,
            "datasource_type": datasource_type.metadata.asset_id,
        }
        if description:
            data["description"] = description
        if properties:
            data["properties"] = properties
        params = {
            "project_id": project.metadata.guid,
            "test": test,
        }
        response = project._platform._connections_api.add_connection(data=json.dumps(data), params=params)
        return Connection(platform=project._platform, project=project, **response.json())

    def model_dump(self, *, by_alias: bool = True, exclude_unset: bool = True, **kwargs: dict) -> dict:
        """Changing default parameters of model_dump to make sure that serialized json math API response.

        Args:
            by_alias: Whether to use alias names in serialization.
            exclude_unset: Whether to exclude unset fields from serialization.
            **kwargs: Additional keyword arguments to pass to the model_dump method.

        Returns:
           A dictionary representation of the model.
        """
        return super().model_dump(exclude_unset=exclude_unset, by_alias=by_alias, **kwargs)

    @property
    def connection_id(self) -> str:
        """Returns id of connection."""
        return self.metadata.asset_id

    @property
    def origin(self) -> dict:
        """Returns origin model dump."""
        return self._origin

    @property
    def actions(self) -> list[DatasourceTypeAction]:
        """Get all supported actions for the connection.

        Returns:
            Retrieved list of Actions.
        """
        params = {
            "project_id": self._project.metadata.guid,
        }
        actions = self._platform._connections_api.get_connection_actions(
            connection_id=self.metadata.asset_id,
            params=params,
        )
        return [
            DatasourceTypeAction(platform=self._platform, project=self, connection=self, **action)
            for action in actions.json().get("actions", list())
        ]

    def _delete(self) -> requests.Response:
        """Remove the Connection.

        Returns:
            A HTTP response.
        """
        params = {
            "project_id": self._project.metadata.guid,
        }
        return self._platform._connections_api.delete_connection(connection_id=self.metadata.asset_id, params=params)

    def _update(self, test: bool = True) -> requests.Response:
        """Update the Connection.

        Args:
            test: whether to test the connection before saving it.
                  If true and validation cannot be estabilished, connection will not be saved.

        Returns:
            A HTTP response.
        """
        params = {
            "project_id": self._project.metadata.guid,
            "test": test,
        }
        data = prepare_json_patch_payload(self.origin["entity"], self.model_dump()["entity"])
        return self._platform._connections_api.update_connection(
            connection_id=self.metadata.asset_id,
            params=params,
            data=data,
        )

    def _copy(self) -> Connection:
        """Update the Connection.

        Returns:
            Copied Connection object.
        """
        params = {
            "project_id": self._project.metadata.guid,
            "target_project_id": self._project.metadata.guid,
        }
        new_connection = self._platform._connections_api.copy_connection(
            connection_id=self.metadata.asset_id,
            params=params,
        )
        return Connection(platform=self._platform, project=self._project, **new_connection.json())


class Connections(CollectionModel):
    """Collection of Connection instances."""

    def __init__(self, platform: Platform, project: Project) -> None:
        """The __init__ of the Connection class.

        Args:
            platform: The Platform object.
            project: Instance of Project in which connection was created.
        """
        super().__init__(platform)
        self.unique_id = "metadata.asset_id"
        self._project = project

    @override
    def __len__(self) -> int:
        connections_json = self._platform._connections_api.get_connections(
            params={
                "project_id": self._project.metadata.guid,
                "limit": 1,
            },
        ).json()
        return connections_json["total_count"]

    def _request_parameters(self) -> list:
        request_params = ["project_id", "connection_id"]
        content_string = self._platform._connections_api.get_swagger().text
        request_path = f"/{self._platform._connections_api.url_path_connections}"
        request_params.extend(get_params_from_swagger(content_string=content_string, request_path=request_path))
        return request_params

    def _get_results_from_api(self, request_params: dict = None, **kwargs: dict) -> CollectionModelResults:
        """Returns results of an api request."""
        request_params_defaults = {
            "project_id": self._project.metadata.guid,
            "limit": 100,
        }
        request_params_unioned: dict[str, Any] = request_params_defaults

        request_params_unioned.update(request_params)

        if "creator" in request_params_unioned:
            request_params_unioned["metadata.creator"] = request_params_unioned["creator"]
            del request_params_unioned["creator"]
        if "name" in request_params_unioned:
            request_params_unioned["entity.name"] = request_params_unioned["name"]
            del request_params_unioned["name"]
        if "datasource_type" in request_params_unioned:
            request_params_unioned["entity.datasource_type"] = request_params_unioned["datasource_type"]
            del request_params_unioned["datasource_type"]
        if "data_source_definition_asset_id" in request_params_unioned:
            request_params_unioned["entity.data_source_definition_asset_id"] = request_params_unioned[
                "data_source_definition_asset_id"
            ]
            del request_params_unioned["data_source_definition_asset_id"]
        if "context" in request_params_unioned:
            request_params_unioned["entity.context"] = request_params_unioned["context"]
            del request_params_unioned["context"]
        if "properties" in request_params_unioned:
            request_params_unioned["entity.properties"] = json.dumps(request_params_unioned["properties"])
        if "flags" in request_params_unioned:
            request_params_unioned["entity.flags"] = request_params_unioned["flags"]
            del request_params_unioned["flags"]

        response = dict()
        if "connection_id" in request_params:
            connection_json = self._platform._connections_api.get_connection(
                connection_id=request_params["connection_id"],
                params={
                    "project_id": self._project.metadata.guid,
                },
            ).json()
            response["resources"] = [connection_json]
        else:
            response = self._platform._connections_api.get_connections(
                params={k: v for k, v in request_params_unioned.items() if v is not None}
            ).json()

            if "next" in response:
                response["next"] = urllib.parse.parse_qs(urllib3.util.parse_url(response["next"]["href"]).query).get(
                    "start"
                )

        return CollectionModelResults(
            results=response,
            class_type=Connection,
            response_bookmark="next",
            request_bookmark="start",
            response_location="resources",
            constructor_params={"platform": self._platform, "project": self._project},
        )


class DiscoveredAsset(BaseModel):
    description: str | None = Field(None, description="A description of the asset.", repr=False)
    details: dict[str, dict[str, Any]] | None = Field(None, repr=False)
    has_children: bool | None = Field(
        None,
        description="True if it is known that the asset has children. False if it is known that the asset does not have children. If it is not known, or it is too expensive to determine this, then this property will not be returned.",
        repr=False,
    )
    id: str | None = Field(None, description="An ID for the asset.", repr=False)
    name: str | None = Field(None, description="A name for the asset.", repr=True)
    path: str | None = Field(
        None, description="The path for the object which can be used to discover child assets.", repr=False
    )
    tags: list[str] | None = Field(None, description="Tags associated with the asset.", repr=False)
    type: str | None = Field(
        None, description="The type of the asset, such as SCHEMA, TABLE, FILE, or FOLDER.", repr=True
    )

    model_config = ConfigDict(frozen=True)


class DatasourceTypeActionProperties(BaseModel):
    input: list[DatasourceTypeProperty] | None = Field(None, description="The input properties.", repr=False)
    output: list[DatasourceTypeProperty] | None = Field(
        None, description="The properties of the action result.", repr=False
    )

    model_config = ConfigDict(frozen=True)


class DatasourceTypeAction(BaseModel):
    description: str | None = Field(None, description="A description of the action.", repr=False)
    name: str | None = Field(None, description="The action name.", repr=True)
    properties: DatasourceTypeActionProperties | None = Field(None, repr=False)

    model_config = ConfigDict(frozen=True)

    def __init__(
        self,
        platform: Platform = None,
        project: Project = None,
        connection: Connection = None,
        **action_json: dict,
    ) -> None:
        """The __init__ of the DatasourceTypeAction class.

        Args:
            action_json: The JSON for the DatasourceTypeAction.
            platform: The Platform object.
            project: The Project object.
            connection: The Connection object.
        """
        super().__init__(**action_json)
        self._platform = platform
        self._project = project
        self._connection = connection

    def model_dump(self, *, by_alias: bool = True, exclude_unset: bool = True, **kwargs: dict) -> dict:
        """Changing default parameters of model_dump to make sure that serialized json math API response.

        Args:
            by_alias: Whether to use alias names in serialization.
            exclude_unset: Whether to exclude unset fields from serialization.
            **kwargs: Additional keyword arguments to pass to the model_dump method.

        Returns:
           A dictionary representation of the model.
        """
        return super().model_dump(exclude_unset=exclude_unset, by_alias=by_alias, **kwargs)

    def execute(
        self, configuration: dict[str, Any] | None = None, connection: Connection | None = None
    ) -> requests.Response:
        """Execute DatasourceTypeAction.

        Args:
            configuration: action config parameters.
            connection: connection to be used.

        Returns:
            A HTTP response.

        Raises:
            ValueError: If no connection is provided.
        """
        if self._connection is None and connection is None:
            raise ValueError("No connection provided.")

        params = {
            "project_id": self._project.metadata.guid,
        }
        data = configuration or dict()
        response = self._platform._connections_api.perform_connection_action(
            connection_id=connection.metadata.asset_id if connection else self._connection.metadata.asset_id,
            action_name=self.name,
            params=params,
            data=data,
        )
        return response


class DatasourceTypeProperties(BaseModel):
    connection: list[DatasourceTypeProperty] = Field(
        default_factory=list, description="The connection properties.", repr=False
    )
    filter: list[DatasourceTypeProperty] | None = Field(
        None, description="The filter properties that can be set for a discovery interaction.", repr=False
    )
    source: list[DatasourceTypeProperty] | None = Field(
        None, description="The properties that can be set for a source interaction.", repr=False
    )
    target: list[DatasourceTypeProperty] | None = Field(
        None, description="The properties that can be set for a target interaction.", repr=False
    )

    model_config = ConfigDict(frozen=True)


class DatasourceType(BaseModel):
    EXPOSED_DATA_PATH: ClassVar[dict] = {"entity": {}}
    actions: list[DatasourceTypeAction] | None = Field(
        None, description="The actions supported for the data source.", repr=False
    )
    allowed_as_source: bool | None = Field(
        None,
        description="Whether the data source can be accessed as a source of data. That is, data can be read from the data source.",
        repr=False,
    )
    allowed_as_target: bool | None = Field(
        None,
        description="Whether the data source can be accessed as a target. That is, data can be written to the data source.",
        repr=False,
    )
    child_source_systems: list[dict[str, dict[str, Any]]] | None = Field(None, repr=False)
    description: str | None = Field(
        None, description="A localized, displayable description of the data source.", repr=False
    )
    label: str | None = Field(None, description='A localized, displayable label such as, "IBM dashDB".', repr=False)
    metadata: ConnectionMetadata | None = Field(None, repr=False)
    name: str | None = Field(None, description='A unique name, such as "dashdb".', repr=True)
    origin_country: str | None = Field(
        None, description="Country which data originated from. - ISO 3166 Country Codes.", repr=False
    )
    owner_id: str | None = Field(
        None,
        description="Owner or creator of connection.  Provided when a service ID token is used to create connection.",
        repr=False,
    )
    properties: DatasourceTypeProperties = Field(default_factory=DatasourceTypeProperties, repr=False)
    status: str | None = Field(None, description="The status of the data source.", repr=False)
    tags: list[str] | None = Field(None, description="Tags associated with a data source type.", repr=False)

    model_config = ConfigDict(frozen=True)

    def __init__(self, platform: Platform = None, **datasource_json: dict) -> None:
        """The __init__ of the Datasource class.

        Args:
            datasource_json: The JSON for the Datasource.
            platform: The Platform object.
        """
        super().__init__(**datasource_json)
        self._platform = platform

    def model_dump(self, *, by_alias: bool = True, exclude_unset: bool = True, **kwargs: dict) -> dict:
        """Changing default parameters of model_dump to make sure that serialized json math API response.

        Args:
            by_alias: Whether to use alias names in serialization.
            exclude_unset: Whether to exclude unset fields from serialization.
            **kwargs: Additional keyword arguments to pass to the model_dump method.

        Returns:
           A dictionary representation of the model.
        """
        return super().model_dump(exclude_unset=exclude_unset, by_alias=by_alias, **kwargs)

    @property
    def datasource_id(self) -> str:
        """Returns id of datasource."""
        return self.metadata.asset_id

    @property
    def required_connection_properties(self) -> list[DatasourceTypeProperty]:
        """Get all required connection properties."""
        return [prty for prty in self.properties.connection if prty.required]


class DatasourceTypes(CollectionModel):
    """Collection of DatasourceType instances."""

    def __init__(self, platform: Platform) -> None:
        """The __init__ of the DatasourceType class.

        Args:
            platform: The Platform object.
        """
        super().__init__(platform)
        self.unique_id = "metadata.asset_id"

    @override
    def __len__(self) -> int:
        datasources_json = self._platform._connections_api.get_datasources(
            params={
                "limit": 1,
            }
        ).json()
        return datasources_json["total_count"]

    def _request_parameters(self) -> list:
        request_params = []
        content_string = self._platform._connections_api.get_swagger().text
        request_path = f"/{self._platform._connections_api.url_path_datasource_types}"
        request_params.extend(get_params_from_swagger(content_string=content_string, request_path=request_path))
        return request_params

    def _get_results_from_api(self, request_params: dict = None, **kwargs: dict) -> CollectionModelResults:
        """Returns results of an api request."""
        request_params_defaults = {
            "limit": 100,
            "offset": 0,
        }
        request_params_unioned: dict[str, Any] = request_params_defaults

        request_params_unioned.update(request_params)

        if "environment" in request_params_unioned:
            request_params_unioned["entity.environment"] = request_params_unioned["environment"]
            del request_params_unioned["environment"]
        if "perspective" in request_params_unioned:
            request_params_unioned["entity.perspective"] = request_params_unioned["perspective"]
            del request_params_unioned["perspective"]
        if "product" in request_params_unioned:
            request_params_unioned["entity.product"] = request_params_unioned["product"]
            del request_params_unioned["product"]

        response = self._platform._connections_api.get_datasources(
            params={k: v for k, v in request_params_unioned.items() if v is not None}
        ).json()

        if "next" in response:
            response["next"] = urllib.parse.parse_qs(urllib3.util.parse_url(response["next"]["href"]).query).get(
                "offset"
            )

        return CollectionModelResults(
            results=response,
            class_type=DatasourceType,
            response_bookmark="next",
            request_bookmark="offset",
            response_location="resources",
            constructor_params={"platform": self._platform},
        )

    @override
    def _paginate(self, **kwargs: dict) -> BaseModel:
        datasource_id = kwargs.pop("datasource_id", None)

        # to allow filter by datasource_id on DatasourceTypes
        # | endpoint dont support this param and it is a property (no way to set repr field)
        for ds in super()._paginate(**kwargs):
            if datasource_id is None or datasource_id == ds.datasource_id:
                yield ds


class ConnectionFile(BaseModel):
    created_at: str | None = Field(None, alias="createdAt", description="Date of creation of the file.", repr=False)
    url: str | None = Field(None, description="Signed URL of the file.", repr=False)
    hash: str | None = Field(None, description="Hash of the file.", repr=False)
    digest: str | None = Field(None, description="Digest of the file.", repr=False)
    file_name: str | None = Field(
        None, alias="fileName", description="Name of the file to be used by connections.", repr=True
    )

    model_config = ConfigDict(frozen=True)

    def __init__(self, platform: Platform = None, **file_json: dict) -> None:
        """The __init__ of the Datasource class.

        Args:
            file_json: The JSON for the ConnectionFile.
            platform: The Platform object.
        """
        super().__init__(**file_json)
        self._platform = platform

    def download(self, output: Path) -> requests.Response:
        """Download a file.

        Args:
            output: destination file path.

        Returns:
            A HTTP response.
        """
        return self._platform._connections_api.get_file(file_name=self.file_name, hash=self.hash, output=output)

    def _delete(self) -> requests.Response:
        """Delete a file.

        Returns:
            A HTTP response.
        """
        return self._platform._connections_api.delete_file(file_name=self.file_name, hash=self.hash)


class ConnectionFiles(CollectionModel):
    """Collection of ConnectionFile instances."""

    def __init__(self, platform: Platform) -> None:
        """The __init__ of the ConnectionFile class.

        Args:
            platform: The Platform object.
        """
        super().__init__(platform)
        self.unique_id = "hash"

    @override
    def __len__(self) -> int:
        datasources_json = self._platform._connections_api.get_file_list(
            params={
                "limit": 0,
            }
        ).json()
        return datasources_json["connection_action_response"]["value"]["total_count"]

    def _request_parameters(self) -> list:
        request_params = []
        content_string = self._platform._connections_api.get_swagger().text
        request_path = f"/{self._platform._connections_api.url_path_connections}/files"
        request_params.extend(get_params_from_swagger(content_string=content_string, request_path=request_path))
        return request_params

    def _get_results_from_api(self, request_params: dict = None, **kwargs: dict) -> CollectionModelResults:
        """Returns results of an api request."""
        request_params_defaults = {
            "limit": 100,
        }
        request_params_unioned: dict[str, Any] = request_params_defaults
        request_params_unioned.update(request_params)

        response = self._platform._connections_api.get_file_list(
            params={k: v for k, v in request_params_unioned.items() if v is not None}
        ).json()
        return CollectionModelResults(
            results=response["connection_action_response"]["value"],
            class_type=ConnectionFile,
            response_bookmark="",
            request_bookmark="",
            response_location="resources",
            constructor_params={"platform": self._platform},
        )



import warnings
from ibm_watsonx_data_integration.services.datastage.models.base_stage import BaseStage
from ibm_watsonx_data_integration.services.datastage.models.enums import XML_OUTPUT
from pydantic import Field
from typing import ClassVar


class xml_output(BaseStage):

    op_name: ClassVar[str] = "PxXMLOutput"
    node_type: ClassVar[str] = "execution_node"
    image: ClassVar[str] = "/data-intg/flows/graphics/palette/XMLOutputPX.svg"
    label: ClassVar[str] = "XML Output"
    runtime_column_propagation: bool = Field(False, alias="runtime_column_propagation")
    auto_column_propagation: bool | None = Field(None, alias="auto_column_propagation")
    buf_free_run: int | None = Field(50, alias="buf_free_run")
    buf_free_run_ronly: int | None = Field(50, alias="buf_free_run_ronly")
    buf_mode: XML_OUTPUT.BufMode | None = Field(XML_OUTPUT.BufMode.default, alias="buf_mode")
    buf_mode_ronly: XML_OUTPUT.BufModeRonly | None = Field(XML_OUTPUT.BufModeRonly.default, alias="buf_mode_ronly")
    coll_type: XML_OUTPUT.CollType | None = Field(XML_OUTPUT.CollType.auto, alias="coll_type")
    combinability: XML_OUTPUT.Combinability | None = Field(XML_OUTPUT.Combinability.auto, alias="combinability")
    current_output_link_type: str = Field("PRIMARY", alias="currentOutputLinkType")
    discard_empty_values: bool | None = Field(False, alias="discard_empty_values")
    discard_empty_values_output: bool | None = Field(False, alias="discard_empty_values_output")
    disk_write_inc: int | None = Field(1048576, alias="disk_write_inc")
    disk_write_inc_ronly: int | None = Field(1048576, alias="disk_write_inc_ronly")
    document_type: str | None = Field("", alias="document_type")
    document_type_output: str | None = Field("", alias="document_type_output")
    document_type_text: str | None = Field("text", alias="document_type_text")
    document_type_text_output: str | None = Field("text", alias="document_type_text_output")
    empty_element_style: str | None = Field("use single tag", alias="empty_element_style")
    empty_element_style_output: str | None = Field("use single tag", alias="empty_element_style_output")
    enable_flow_acp_control: bool = Field(True, alias="enableFlowAcpControl")
    enable_grammar_caching: bool | None = Field(False, alias="enable_grammar_caching")
    enable_schemaless_design: bool = Field(False, alias="enableSchemalessDesign")
    execmode: XML_OUTPUT.Execmode | None = Field(XML_OUTPUT.Execmode.default_seq, alias="execmode")
    flow_dirty: str | None = Field("false", alias="flow_dirty")
    generate_comment: bool | None = Field(True, alias="generate_comment")
    generate_comment_output: bool | None = Field(True, alias="generate_comment_output")
    generate_full_document: bool | None = Field(False, alias="generate_full_document")
    generate_full_document_output: bool | None = Field(False, alias="generate_full_document_output")
    header: str | None = Field("", alias="header")
    header_output: str | None = Field("", alias="header_output")
    header_text: str | None = Field("text", alias="header_text")
    header_text_output: str | None = Field("text", alias="header_text_output")
    heap_size: int | None = Field(256, alias="heap_size")
    hide: bool | None = Field(False, alias="hide")
    include_document: bool | None = Field(False, alias="include_document")
    include_document_output: bool | None = Field(False, alias="include_document_output")
    include_header: bool | None = Field(False, alias="include_header")
    include_header_output: bool | None = Field(False, alias="include_header_output")
    include_namespaces: bool | None = Field(False, alias="include_namespaces")
    include_namespaces_output: bool | None = Field(False, alias="include_namespaces_output")
    include_xml_chunk: bool | None = Field(False, alias="include_xml_chunk")
    include_xml_chunk_output: bool | None = Field(False, alias="include_xml_chunk_output")
    indentation_character: str | None = Field("\\s", alias="indentation_character")
    indentation_character_output: str | None = Field("\\s", alias="indentation_character_output")
    indentation_length: int | None = Field(2, alias="indentation_length")
    indentation_length_output: int | None = Field(2, alias="indentation_length_output")
    inherit_properties: bool | None = Field(False, alias="inherit_properties")
    input_count: int | None = Field(0, alias="input_count")
    input_link_description: list | None = Field("", alias="inputLinkDescription")
    inputcol_properties: list | None = Field([], alias="inputcolProperties")
    is_icp4d: bool | None = Field(False, alias="isIcp4d")
    is_reject_link: bool | None = Field(False, alias="is_reject_link")
    is_remote_engine: bool | None = Field(False, alias="isRemoteEngine")
    key_col_select: XML_OUTPUT.KeyColSelect | None = Field(XML_OUTPUT.KeyColSelect.default, alias="keyColSelect")
    key_cols_coll: list | None = Field([], alias="keyColsColl")
    key_cols_coll_ordered: list | None = Field([], alias="keyColsCollOrdered")
    key_cols_coll_robin: list | None = Field([], alias="keyColsCollRobin")
    key_cols_none: list | None = Field([], alias="keyColsNone")
    key_cols_part: list | None = Field([], alias="keyColsPart")
    log_reject_errors: bool | None = Field(False, alias="log_reject_errors")
    max_mem_buf_size: int | None = Field(3145728, alias="max_mem_buf_size")
    max_mem_buf_size_ronly: int | None = Field(3145728, alias="max_mem_buf_size_ronly")
    multi_line_output: bool | None = Field(False, alias="multi_line_output")
    multi_line_output_output: bool | None = Field(False, alias="multi_line_output_output")
    namespace_declaration: str | None = Field("", alias="namespace_declaration")
    namespace_declaration_output: str | None = Field("", alias="namespace_declaration_output")
    new_line_style: str | None = Field("MAC", alias="new_line_style")
    new_line_style_output: str | None = Field("MAC", alias="new_line_style_output")
    other_options: str | None = Field(None, alias="other_options")
    output_acp_should_hide: bool = Field(True, alias="outputAcpShouldHide")
    output_count: int | None = Field(0, alias="output_count")
    output_file: str | None = Field("", alias="output_file")
    output_file_output: str | None = Field("", alias="output_file_output")
    output_link_description: list | None = Field("", alias="outputLinkDescription")
    output_row_generation: str | None = Field("use column", alias="output_row_generation")
    output_row_generation_output: str | None = Field("use column", alias="output_row_generation_output")
    output_row_trigger_column: str | None = Field("", alias="output_row_trigger_column")
    output_row_trigger_column_output: str | None = Field("", alias="output_row_trigger_column_output")
    outputcol_properties: list | None = Field([], alias="outputcolProperties")
    part_client_dbname: str | None = Field(None, alias="part_client_dbname")
    part_client_instance: str | None = Field(None, alias="part_client_instance")
    part_dbconnection: str | None = Field("", alias="part_dbconnection")
    part_stable: bool | None = Field(None, alias="part_stable")
    part_stable_coll: bool | None = Field(False, alias="part_stable_coll")
    part_stable_ordered: bool | None = Field(False, alias="part_stable_ordered")
    part_stable_roundrobin_coll: bool | None = Field(False, alias="part_stable_roundrobin_coll")
    part_table: str | None = Field(None, alias="part_table")
    part_type: XML_OUTPUT.PartType | None = Field(XML_OUTPUT.PartType.auto, alias="part_type")
    part_unique: bool | None = Field(None, alias="part_unique")
    part_unique_coll: bool | None = Field(False, alias="part_unique_coll")
    part_unique_ordered: bool | None = Field(False, alias="part_unique_ordered")
    part_unique_roundrobin_coll: bool | None = Field(False, alias="part_unique_roundrobin_coll")
    perform_sort: bool | None = Field(False, alias="perform_sort")
    perform_sort_coll: bool | None = Field(False, alias="perform_sort_coll")
    perform_sort_modulus: bool | None = Field(False, alias="perform_sort_modulus")
    preserve: XML_OUTPUT.Preserve | None = Field(XML_OUTPUT.Preserve.default_propagate, alias="preserve")
    queue_upper_size: int | None = Field(0, alias="queue_upper_size")
    queue_upper_size_ronly: int | None = Field(0, alias="queue_upper_size_ronly")
    reject_message_column: str | None = Field("", alias="reject_message_column")
    replace_nulls: bool | None = Field(False, alias="replace_nulls")
    replace_nulls_output: bool | None = Field(False, alias="replace_nulls_output")
    runtime_column_propagation: bool | None = Field(None, alias="runtime_column_propagation")
    show_coll_type: int | None = Field(0, alias="showCollType")
    show_part_type: int | None = Field(1, alias="showPartType")
    show_sort_options: int | None = Field(0, alias="showSortOptions")
    sort_instructions: str | None = Field("", alias="sortInstructions")
    sort_instructions_text: str | None = Field("", alias="sortInstructionsText")
    stage_description: list | None = Field("", alias="stageDescription")
    validate_output: bool | None = Field(False, alias="validate_output")
    write_to_file: bool | None = Field(False, alias="write_to_file")
    write_to_file_output: bool | None = Field(False, alias="write_to_file_output")
    xml_chunk: str | None = Field("", alias="xml_chunk")
    xml_chunk_output: str | None = Field("", alias="xml_chunk_output")
    xml_chunk_text: str | None = Field("text", alias="xml_chunk_text")
    xml_chunk_text_output: str | None = Field("text", alias="xml_chunk_text_output")
    xml_encoding: str | None = Field("UTF-8", alias="xml_encoding")
    xml_encoding_output: str | None = Field("UTF-8", alias="xml_encoding_output")
    xml_error_mapping: XML_OUTPUT.XmlErrorMapping | None = Field(
        XML_OUTPUT.XmlErrorMapping.DS_FATAL, alias="xml_error_mapping"
    )
    xml_fatal_mapping: XML_OUTPUT.XmlFatalMapping | None = Field(
        XML_OUTPUT.XmlFatalMapping.DS_REJECT, alias="xml_fatal_mapping"
    )
    xml_validation_level: XML_OUTPUT.XmlValidationLevel | None = Field(
        XML_OUTPUT.XmlValidationLevel.default, alias="xml_validation_level"
    )
    xml_warning_mapping: XML_OUTPUT.XmlWarningMapping | None = Field(
        XML_OUTPUT.XmlWarningMapping.DS_WARNING, alias="xml_warning_mapping"
    )

    def validate_parameters(self) -> tuple[set, set]:
        include = set()
        exclude = set()

        include.add("preserve") if (self.output_count and self.output_count > 0) else exclude.add("preserve")
        include.add("max_mem_buf_size") if (self.buf_mode != "nobuffer") else exclude.add("max_mem_buf_size")
        include.add("buf_free_run") if (self.buf_mode != "nobuffer") else exclude.add("buf_free_run")
        include.add("queue_upper_size") if (self.buf_mode != "nobuffer") else exclude.add("queue_upper_size")
        include.add("disk_write_inc") if (self.buf_mode != "nobuffer") else exclude.add("disk_write_inc")

        (
            include.add("runtime_column_propagation")
            if (not self.enable_schemaless_design)
            else exclude.add("runtime_column_propagation")
        )
        (
            include.add("auto_column_propagation")
            if (not self.output_acp_should_hide)
            else exclude.add("auto_column_propagation")
        )
        (
            include.add("max_mem_buf_size_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("max_mem_buf_size_ronly")
        )
        (
            include.add("buf_free_run_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("buf_free_run_ronly")
        )
        (
            include.add("queue_upper_size_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("queue_upper_size_ronly")
        )
        (
            include.add("disk_write_inc_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("disk_write_inc_ronly")
        )
        (
            include.add("part_stable")
            if (
                ((self.show_part_type))
                and (self.part_type != "auto")
                and (self.part_type != "db2connector")
                and (self.show_sort_options)
            )
            else exclude.add("part_stable")
        )
        (
            include.add("part_unique")
            if (
                ((self.show_part_type))
                and (self.part_type != "auto")
                and (self.part_type != "db2connector")
                and (self.show_sort_options)
            )
            else exclude.add("part_unique")
        )
        (
            include.add("key_cols_part")
            if (
                (
                    ((self.show_part_type))
                    and ((not self.show_coll_type))
                    and ((self.part_type != "auto"))
                    and ((self.part_type != "db2connector"))
                    and ((self.part_type != "modulus"))
                )
                or (
                    ((self.show_part_type))
                    and (not self.show_coll_type)
                    and (self.part_type == "modulus")
                    and (self.perform_sort_modulus)
                )
            )
            else exclude.add("key_cols_part")
        )
        (
            include.add("part_dbconnection")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_dbconnection")
        )
        (
            include.add("part_client_dbname")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_client_dbname")
        )
        (
            include.add("part_client_instance")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_client_instance")
        )
        (
            include.add("part_table")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_table")
        )
        (
            include.add("perform_sort")
            if (((self.show_part_type)) and (((self.part_type == "hash")) or (self.part_type == "range")))
            else exclude.add("perform_sort")
        )
        (
            include.add("perform_sort_modulus")
            if (((self.show_part_type)) and (self.part_type == "modulus"))
            else exclude.add("perform_sort_modulus")
        )
        (
            include.add("key_col_select")
            if (
                ((self.show_part_type))
                and (self.part_type == "modulus")
                and (not self.perform_sort_modulus)
            )
            else exclude.add("key_col_select")
        )
        (
            include.add("sort_instructions")
            if (
                ((self.show_part_type))
                and (
                    ((self.part_type == "db2part"))
                    or (self.part_type == "entire")
                    or (self.part_type == "random")
                    or (self.part_type == "roundrobin")
                    or (self.part_type == "same")
                )
            )
            else exclude.add("sort_instructions")
        )
        (
            include.add("sort_instructions_text")
            if (
                ((self.show_part_type))
                and (
                    ((self.part_type == "db2part"))
                    or (self.part_type == "entire")
                    or (self.part_type == "random")
                    or (self.part_type == "roundrobin")
                    or (self.part_type == "same")
                )
            )
            else exclude.add("sort_instructions_text")
        )
        include.add("coll_type") if (self.show_coll_type) else exclude.add("coll_type")
        include.add("part_type") if (self.show_part_type) else exclude.add("part_type")
        (
            include.add("perform_sort_coll")
            if (
                (
                    ((self.show_coll_type))
                    and (
                        ((self.coll_type == "ordered"))
                        or ((self.coll_type == "roundrobin_coll"))
                        or ((self.coll_type == "sortmerge"))
                    )
                )
                or (((not self.show_part_type)) and (not self.show_coll_type))
            )
            else exclude.add("perform_sort_coll")
        )
        (
            include.add("key_cols_coll")
            if (
                ((self.show_coll_type))
                and (not self.show_part_type)
                and (self.coll_type != "auto")
                and (((self.coll_type == "sortmerge")) or (self.perform_sort_coll))
            )
            else exclude.add("key_cols_coll")
        )
        (
            include.add("key_cols_none")
            if (
                ((not self.show_part_type))
                and (not self.show_coll_type)
                and (self.perform_sort_coll)
            )
            else exclude.add("key_cols_none")
        )
        (
            include.add("part_stable_coll")
            if (
                (((self.perform_sort_coll)))
                and (
                    (
                        ((not self.show_part_type))
                        and ((self.show_coll_type))
                        and ((self.coll_type != "auto"))
                        and ((self.show_sort_options))
                    )
                    or (
                        ((not self.show_part_type))
                        and (not self.show_coll_type)
                        and (self.show_sort_options)
                    )
                )
            )
            else exclude.add("part_stable_coll")
        )
        (
            include.add("part_unique_coll")
            if (
                (((self.perform_sort_coll)))
                and (
                    (
                        ((not self.show_part_type))
                        and ((self.show_coll_type))
                        and ((self.coll_type != "auto"))
                        and ((self.show_sort_options))
                    )
                    or (
                        ((not self.show_part_type))
                        and (not self.show_coll_type)
                        and (self.show_sort_options)
                    )
                )
            )
            else exclude.add("part_unique_coll")
        )
        return include, exclude

    def get_input_ports_props(self) -> dict:
        include, exclude = self.validate()
        props = {"runtime_column_propagation"}
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_input_cardinality(self) -> dict:
        return {"min": 1, "max": 1}

    def get_output_ports_props(self) -> dict:
        include, exclude = self.validate()
        props = {
            "discard_empty_values_output",
            "document_type_output",
            "document_type_text_output",
            "empty_element_style_output",
            "generate_comment_output",
            "generate_full_document_output",
            "header_output",
            "header_text_output",
            "include_document_output",
            "include_header_output",
            "include_namespaces_output",
            "include_xml_chunk_output",
            "indentation_character_output",
            "indentation_length_output",
            "inherit_properties",
            "is_reject_link",
            "multi_line_output_output",
            "namespace_declaration_output",
            "new_line_style_output",
            "output_file_output",
            "output_row_generation_output",
            "output_row_trigger_column_output",
            "reject_message_column",
            "replace_nulls_output",
            "xml_chunk_output",
            "xml_chunk_text_output",
            "xml_encoding_output",
        }
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_output_cardinality(self) -> dict:
        return {"min": 0, "max": 2}

    def get_parameters_props(self) -> dict:
        include, exclude = self.validate()
        props = {
            "auto_column_propagation",
            "buf_free_run",
            "buf_free_run_ronly",
            "buf_mode",
            "buf_mode_ronly",
            "coll_type",
            "combinability",
            "current_output_link_type",
            "discard_empty_values",
            "discard_empty_values_output",
            "disk_write_inc",
            "disk_write_inc_ronly",
            "document_type",
            "document_type_output",
            "document_type_text",
            "document_type_text_output",
            "empty_element_style",
            "empty_element_style_output",
            "enable_flow_acp_control",
            "enable_grammar_caching",
            "enable_schemaless_design",
            "execmode",
            "flow_dirty",
            "generate_comment",
            "generate_comment_output",
            "generate_full_document",
            "generate_full_document_output",
            "header",
            "header_output",
            "header_text",
            "header_text_output",
            "heap_size",
            "hide",
            "include_document",
            "include_document_output",
            "include_header",
            "include_header_output",
            "include_namespaces",
            "include_namespaces_output",
            "include_xml_chunk",
            "include_xml_chunk_output",
            "indentation_character",
            "indentation_character_output",
            "indentation_length",
            "indentation_length_output",
            "inherit_properties",
            "input_count",
            "input_link_description",
            "inputcol_properties",
            "is_icp4d",
            "is_reject_link",
            "is_remote_engine",
            "key_col_select",
            "key_cols_coll",
            "key_cols_coll_ordered",
            "key_cols_coll_robin",
            "key_cols_none",
            "key_cols_part",
            "log_reject_errors",
            "max_mem_buf_size",
            "max_mem_buf_size_ronly",
            "multi_line_output",
            "multi_line_output_output",
            "namespace_declaration",
            "namespace_declaration_output",
            "new_line_style",
            "new_line_style_output",
            "other_options",
            "output_acp_should_hide",
            "output_count",
            "output_file",
            "output_file_output",
            "output_link_description",
            "output_row_generation",
            "output_row_generation_output",
            "output_row_trigger_column",
            "output_row_trigger_column_output",
            "outputcol_properties",
            "part_client_dbname",
            "part_client_instance",
            "part_dbconnection",
            "part_stable",
            "part_stable_coll",
            "part_stable_ordered",
            "part_stable_roundrobin_coll",
            "part_table",
            "part_type",
            "part_unique",
            "part_unique_coll",
            "part_unique_ordered",
            "part_unique_roundrobin_coll",
            "perform_sort",
            "perform_sort_coll",
            "perform_sort_modulus",
            "preserve",
            "queue_upper_size",
            "queue_upper_size_ronly",
            "reject_message_column",
            "replace_nulls",
            "replace_nulls_output",
            "runtime_column_propagation",
            "show_coll_type",
            "show_part_type",
            "show_sort_options",
            "sort_instructions",
            "sort_instructions_text",
            "stage_description",
            "validate_output",
            "write_to_file",
            "write_to_file_output",
            "xml_chunk",
            "xml_chunk_output",
            "xml_chunk_text",
            "xml_chunk_text_output",
            "xml_encoding",
            "xml_encoding_output",
            "xml_error_mapping",
            "xml_fatal_mapping",
            "xml_validation_level",
            "xml_warning_mapping",
        }
        required = {
            "current_output_link_type",
            "enable_flow_acp_control",
            "enable_schemaless_design",
            "output_acp_should_hide",
        }
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_app_data_props(self) -> dict:
        return {
            "datastage": {
                "maxRejectOutputs": 0,
                "minRejectOutputs": 0,
                "maxReferenceInputs": 0,
                "minReferenceInputs": 0,
            }
        }



from .api import *
from .codegen import *
from .models import *

__all__ = [
    "DataStageFlow",
    "Schema",
    "Field",
    "PythonGenerator",
    "SPLIT_VECTOR",
    "DIFFERENCE",
    "DECODE",
    "SURVIVE",
    "SAMPLE",
    "ADDRESS_VERIFICATION",
    "MATCH_FREQUENCY",
    "ONE_SOURCE_MATCH",
    "SWITCH",
    "WAVE_GENERATOR",
    "COLUMN_GENERATOR",
    "TWO_SOURCE_MATCH",
    "DATASET",
    "TRANSFORMER",
    "JOIN",
    "SORT",
    "MAKE_VECTOR",
    "FILTER",
    "PROMOTE_SUBRECORD",
    "INVESTIGATE",
    "REMOVE_DUPLICATES",
    "FILESET",
    "SURROGATE_KEY_GENERATOR",
    "COLUMN_EXPORT",
    "SLOWLY_CHANGING_DIMENSION",
    "COMPRESS",
    "COLUMN_IMPORT",
    "SAPBAPI",
    "TAIL",
    "EXTERNAL_FILTER",
    "FUNNEL",
    "STANDARDIZE",
    "EXPAND",
    "LOOKUPFILESET",
    "BLOOM_FILTER",
    "HEAD",
    "ROW_GENERATOR",
    "CHANGE_CAPTURE",
    "CHANGE_APPLY",
    "PIVOT",
    "CHECKSUM",
    "WRITE_RANGE_MAP",
    "DATA_RULES",
    "COMBINE_RECORDS",
    "MERGE",
    "COMPARE",
    "AGGREGATOR",
    "SEQUENTIALFILE",
    "COPY",
    "GENERIC",
    "MAKE_SUBRECORD",
    "COMPLEX_FLAT_FILE",
    "SPLIT_SUBRECORD",
    "PEEK",
    "EXTERNAL_TARGET",
    "EXTERNAL_SOURCE",
    "MODIFY",
    "ENCODE",
    "LOOKUP",
    "SALESFORCEAPI",
    "MATCH360",
    "SYBASE",
    "SAPDELTAEXTRACT",
    "NETEZZA",
    "POSTGRESQL",
    "FTP",
    "MONGODB_IBMCLOUD",
    "SAPBULKEXTRACT",
    "DB2CLOUD",
    "PRESTO",
    "AZURE_FILE_STORAGE",
    "ORACLE_DATASTAGE",
    "GENERICS3",
    "BOX",
    "PLANNING_ANALYTICS",
    "DREMIO",
    "GOOGLE_LOOKER",
    "TABLEAU",
    "POSTGRESQL_IBMCLOUD",
    "TERADATA",
    "MARIADB",
    "CLOUD_OBJECT_STORAGE",
    "APACHE_KAFKA",
    "WATSONX_DATA",
    "TERADATA_DATASTAGE",
    "APACHE_HBASE",
    "BIGSQL",
    "ORACLE",
    "AMAZONS3",
    "AZURE_COSMOS",
    "AZURE_DATABRICKS",
    "CASSANDRA_DATASTAGE",
    "HDFS_APACHE",
    "AMAZON_REDSHIFT",
    "SAPIQ",
    "IMPALA",
    "DB2ZOS",
    "BIGQUERY",
    "INFORMIX",
    "IBM_MQ",
    "JDBC",
    "SAPIDOC",
    "DERBY",
    "MINIO",
    "EXASOL",
    "DB2WAREHOUSE",
    "MONGODB",
    "DVM",
    "GOOGLE_CLOUD_STORAGE",
    "ODBC",
    "HTTP",
    "DV",
    "CASSANDRA",
    "AZURE_POSTGRESQL",
    "MYSQL",
    "SQLSERVER",
    "GREENPLUM",
    "AZUREDATALAKE",
    "GOOGLE_PUB_SUB",
    "DENODO",
    "DROPBOX",
    "SALESFORCE",
    "AZURESQL",
    "DB2",
    "STORAGE_VOLUME",
    "NETEZZA_OPTIMIZED",
    "APACHE_HIVE",
    "SNOWFLAKE",
    "MYSQL_COMPOSE",
    "DB2FORDATASTAGE",
    "AMAZONRDS_ORACLE",
    "DB2ISERIES",
    "AZURESYNAPSE",
    "MYSQL_AMAZON",
    "COGNOS_ANALYTICS",
    "VERTICA",
    "ELASTICSEARCH",
    "AMAZON_POSTGRESQL",
    "SAPHANA",
    "SAPODATA",
    "SINGLESTORE",
    "AZURE_BLOB_STORAGE",
    "DATASTAX",
    "HIERARCHICAL_DATA",
    "WEB_SERVICE",
    "XML_OUTPUT",
    "JAVA_INTEGRATION",
    "XML_INPUT",
    "REST",
    "SALESFORCEAPI_CONNECTION",
    "SAPDELTAEXTRACT_CONNECTION",
    "FTP_CONNECTION",
    "MONGODB_IBMCLOUD_CONNECTION",
    "SAPBULKEXTRACT_CONNECTION",
    "AZURE_FILE_STORAGE_CONNECTION",
    "ORACLE_DATASTAGE_CONNECTION",
    "GENERICS3_CONNECTION",
    "PLANNING_ANALYTICS_CONNECTION",
    "DREMIO_CONNECTION",
    "TABLEAU_CONNECTION",
    "TERADATA_CONNECTION",
    "CLOUD_OBJECT_STORAGE_CONNECTION",
    "APACHE_KAFKA_CONNECTION",
    "WATSONX_DATA_CONNECTION",
    "TERADATA_DATASTAGE_CONNECTION",
    "APACHE_HBASE_CONNECTION",
    "ORACLE_CONNECTION",
    "AMAZONS3_CONNECTION",
    "AZURE_COSMOS_CONNECTION",
    "AZURE_DATABRICKS_CONNECTION",
    "CASSANDRA_DATASTAGE_CONNECTION",
    "HDFS_APACHE_CONNECTION",
    "AMAZON_REDSHIFT_CONNECTION",
    "IMPALA_CONNECTION",
    "BIGQUERY_CONNECTION",
    "IBM_MQ_CONNECTION",
    "JDBC_CONNECTION",
    "SAPIDOC_CONNECTION",
    "DB2WAREHOUSE_CONNECTION",
    "MONGODB_CONNECTION",
    "GOOGLE_CLOUD_STORAGE_CONNECTION",
    "ODBC_CONNECTION",
    "DV_CONNECTION",
    "CASSANDRA_CONNECTION",
    "AZURE_POSTGRESQL_CONNECTION",
    "SQLSERVER_CONNECTION",
    "AZUREDATALAKE_CONNECTION",
    "GOOGLE_PUB_SUB_CONNECTION",
    "DROPBOX_CONNECTION",
    "AZURESQL_CONNECTION",
    "DB2_CONNECTION",
    "APACHE_HIVE_CONNECTION",
    "SNOWFLAKE_CONNECTION",
    "DB2FORDATASTAGE_CONNECTION",
    "AMAZONRDS_ORACLE_CONNECTION",
    "DB2ISERIES_CONNECTION",
    "AZURESYNAPSE_CONNECTION",
    "COGNOS_ANALYTICS_CONNECTION",
    "ELASTICSEARCH_CONNECTION",
    "SAPODATA_CONNECTION",
    "AZURE_BLOB_STORAGE_CONNECTION",
    "DATASTAX_CONNECTION",
    "lookup",
    "transformer",
    "complex_flat_file",
    "rest",
    "FIELD",
]



from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from ibm_watsonx_data_integration.services.datastage.models.enums import GOOGLE_CLOUD_STORAGE_CONNECTION
from pydantic import ConfigDict, Field
from typing import ClassVar


class AuthMethod(Enum):
    credentials = "credentials"
    credentials_oauth2 = "credentials_oauth2"
    workload_identity_federation_token = "workload_identity_federation_token"
    workload_identity_federation_token_url = "workload_identity_federation_token_url"


class ProxyProtocol(Enum):
    http = "http"
    https = "https"


class TokenFormat(Enum):
    json = "json"
    text = "text"


class TokenType(Enum):
    aws4_request = "aws4_request"
    access_token = "access_token"
    id_token = "id_token"
    jwt = "jwt"
    saml2 = "saml2"


class TokenUrlMethod(Enum):
    get = "get"
    post = "post"
    put = "put"


class GoogleCloudStorageConn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "05b7f0ea-6ae4-45e2-a455-cc280f110825"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    ds_host: str | None = Field(None, alias="_host")
    ds_port: str | None = Field(None, alias="_port")
    access_token: str = Field(None, alias="access_token")
    authentication_method: GOOGLE_CLOUD_STORAGE_CONNECTION.AuthMethod | None = Field(None, alias="auth_method")
    discover_data_assets: bool | None = Field(None, alias="auto_discovery")
    bucket: str | None = Field(None, alias="bucket")
    das_api_key: str = Field(None, alias="cas_api_key")
    das_endpoint: str = Field(None, alias="cas_endpoint")
    das_instance_id: str = Field(None, alias="cas_instance_id")
    catalog_name: str | None = Field(None, alias="catalog_name")
    client_id: str = Field(None, alias="client_id")
    client_secret: str = Field(None, alias="client_secret")
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    credentials: str = Field(None, alias="credentials")
    credentials_file_path: str = Field(None, alias="credentials_file")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    iceberg_rest_catalog_endpoint: str | None = Field(None, alias="mds_rest_endpoint")
    project_id: str | None = Field(None, alias="project_id")
    proxy: bool | None = Field(False, alias="proxy")
    proxy_host: str = Field(None, alias="proxy_host")
    proxy_password: str | None = Field(None, alias="proxy_password")
    proxy_port: int = Field(None, alias="proxy_port")
    proxy_protocol: GOOGLE_CLOUD_STORAGE_CONNECTION.ProxyProtocol | None = Field(
        GOOGLE_CLOUD_STORAGE_CONNECTION.ProxyProtocol.https, alias="proxy_protocol"
    )
    proxy_username: str | None = Field(None, alias="proxy_user")
    refresh_token: str = Field(None, alias="refresh_token")
    service_account_email: str = Field(None, alias="service_account_email")
    service_account_token_lifetime: int | None = Field(None, alias="service_account_token_lifetime")
    port_is_ssl_enabled: bool | None = Field(False, alias="ssl")
    ssl_certificate: str | None = Field(None, alias="ssl_certificate")
    validate_ssl_certificate: bool | None = Field(None, alias="ssl_certificate_validation")
    security_token_service_audience: str = Field(None, alias="sts_audience")
    token_field_name: str = Field(None, alias="token_field_name")
    token_format: GOOGLE_CLOUD_STORAGE_CONNECTION.TokenFormat | None = Field(
        GOOGLE_CLOUD_STORAGE_CONNECTION.TokenFormat.text, alias="token_format"
    )
    token_type: GOOGLE_CLOUD_STORAGE_CONNECTION.TokenType | None = Field(
        GOOGLE_CLOUD_STORAGE_CONNECTION.TokenType.id_token, alias="token_type"
    )
    token_url: str = Field(None, alias="token_url")
    request_body: str | None = Field(None, alias="token_url_body")
    http_headers: str | None = Field(None, alias="token_url_headers")
    http_method: GOOGLE_CLOUD_STORAGE_CONNECTION.TokenUrlMethod | None = Field(
        GOOGLE_CLOUD_STORAGE_CONNECTION.TokenUrlMethod.get, alias="token_url_method"
    )
    use_watsonx_credential_provider: bool | None = Field(None, alias="use_watsonx_credential_provider")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    ssl_certificate_file: str | None = Field(None, alias="ssl_certificate_file")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        (
            include.add("credentials_file_path")
            if (
                ((not self.credentials))
                and (not self.client_id)
                and (not self.client_secret)
                and (not self.access_token)
                and (not self.refresh_token)
                and (not self.security_token_service_audience)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "credentials_file"
                        )
                        or (self.authentication_method == "credentials_file")
                    )
                )
                and (not self.defer_credentials)
                and (not self.use_watsonx_credential_provider)
            )
            else exclude.add("credentials_file_path")
        )
        (
            include.add("token_field_name")
            if (
                self.token_format
                and (
                    (hasattr(self.token_format, "value") and self.token_format.value == "json")
                    or (self.token_format == "json")
                )
            )
            else exclude.add("token_field_name")
        )
        (
            include.add("credentials")
            if (
                ((not self.credentials_file_path))
                and (not self.client_id)
                and (not self.client_secret)
                and (not self.access_token)
                and (not self.refresh_token)
                and (not self.security_token_service_audience)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "credentials"
                        )
                        or (self.authentication_method == "credentials")
                    )
                )
                and (not self.defer_credentials)
                and (not self.use_watsonx_credential_provider)
            )
            else exclude.add("credentials")
        )
        (
            include.add("http_headers")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.access_token)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "workload_identity_federation_token_url"
                        )
                        or (self.authentication_method == "workload_identity_federation_token_url")
                    )
                )
            )
            else exclude.add("http_headers")
        )
        include.add("proxy_protocol") if (self.proxy) else exclude.add("proxy_protocol")
        (
            include.add("security_token_service_audience")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.client_id)
                and (
                    (
                        (
                            self.authentication_method
                            and (
                                (
                                    hasattr(self.authentication_method, "value")
                                    and self.authentication_method.value == "workload_identity_federation_token"
                                )
                                or (self.authentication_method == "workload_identity_federation_token")
                            )
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "workload_identity_federation_token_url"
                            )
                            or (self.authentication_method == "workload_identity_federation_token_url")
                        )
                    )
                )
            )
            else exclude.add("security_token_service_audience")
        )
        include.add("proxy_password") if (self.proxy) else exclude.add("proxy_password")
        (
            include.add("token_type")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.client_id)
                and (
                    (
                        (
                            self.authentication_method
                            and (
                                (
                                    hasattr(self.authentication_method, "value")
                                    and self.authentication_method.value == "workload_identity_federation_token"
                                )
                                or (self.authentication_method == "workload_identity_federation_token")
                            )
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "workload_identity_federation_token_url"
                            )
                            or (self.authentication_method == "workload_identity_federation_token_url")
                        )
                    )
                )
            )
            else exclude.add("token_type")
        )
        (
            include.add("port_is_ssl_enabled")
            if (
                (
                    ((self.proxy))
                    and (
                        (
                            self.proxy_protocol
                            and (
                                (hasattr(self.proxy_protocol, "value") and self.proxy_protocol.value == "https")
                                or (self.proxy_protocol == "https")
                            )
                        )
                    )
                )
                or (self.use_watsonx_credential_provider)
            )
            else exclude.add("port_is_ssl_enabled")
        )
        (
            include.add("client_id")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.security_token_service_audience)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "credentials_oauth2"
                        )
                        or (self.authentication_method == "credentials_oauth2")
                    )
                )
                and (not self.defer_credentials)
                and (not self.use_watsonx_credential_provider)
            )
            else exclude.add("client_id")
        )
        (
            include.add("token_format")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.client_id)
                and (
                    (
                        (
                            self.authentication_method
                            and (
                                (
                                    hasattr(self.authentication_method, "value")
                                    and self.authentication_method.value == "workload_identity_federation_token"
                                )
                                or (self.authentication_method == "workload_identity_federation_token")
                            )
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "workload_identity_federation_token_url"
                            )
                            or (self.authentication_method == "workload_identity_federation_token_url")
                        )
                    )
                )
            )
            else exclude.add("token_format")
        )
        include.add("proxy_host") if (self.proxy) else exclude.add("proxy_host")
        (
            include.add("request_body")
            if (
                (
                    (
                        self.http_method
                        and (
                            (hasattr(self.http_method, "value") and self.http_method.value == "post")
                            or (self.http_method == "post")
                        )
                    )
                )
                or (
                    self.http_method
                    and (
                        (hasattr(self.http_method, "value") and self.http_method.value == "put")
                        or (self.http_method == "put")
                    )
                )
            )
            else exclude.add("request_body")
        )
        (
            include.add("validate_ssl_certificate")
            if (self.port_is_ssl_enabled)
            else exclude.add("validate_ssl_certificate")
        )
        (
            include.add("client_secret")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.security_token_service_audience)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "credentials_oauth2"
                        )
                        or (self.authentication_method == "credentials_oauth2")
                    )
                )
                and (not self.defer_credentials)
                and (not self.use_watsonx_credential_provider)
            )
            else exclude.add("client_secret")
        )
        (
            include.add("token_url")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.access_token)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "workload_identity_federation_token_url"
                        )
                        or (self.authentication_method == "workload_identity_federation_token_url")
                    )
                )
            )
            else exclude.add("token_url")
        )
        (
            include.add("ssl_certificate")
            if ((((not self.ssl_certificate_file))) and (self.port_is_ssl_enabled))
            else exclude.add("ssl_certificate")
        )
        include.add("proxy_port") if (self.proxy) else exclude.add("proxy_port")
        (
            include.add("http_method")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.access_token)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "workload_identity_federation_token_url"
                        )
                        or (self.authentication_method == "workload_identity_federation_token_url")
                    )
                )
            )
            else exclude.add("http_method")
        )
        include.add("proxy_username") if (self.proxy) else exclude.add("proxy_username")
        (
            include.add("access_token")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.token_url)
                and (
                    (
                        (
                            self.authentication_method
                            and (
                                (
                                    hasattr(self.authentication_method, "value")
                                    and self.authentication_method.value == "credentials_oauth2"
                                )
                                or (self.authentication_method == "credentials_oauth2")
                            )
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "workload_identity_federation_token"
                            )
                            or (self.authentication_method == "workload_identity_federation_token")
                        )
                    )
                )
                and (not self.defer_credentials)
                and (not self.use_watsonx_credential_provider)
            )
            else exclude.add("access_token")
        )
        (
            include.add("refresh_token")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.security_token_service_audience)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "credentials_oauth2"
                        )
                        or (self.authentication_method == "credentials_oauth2")
                    )
                )
                and (not self.defer_credentials)
                and (not self.use_watsonx_credential_provider)
            )
            else exclude.add("refresh_token")
        )
        include.add("das_api_key") if (self.use_watsonx_credential_provider) else exclude.add("das_api_key")
        include.add("das_instance_id") if (self.use_watsonx_credential_provider) else exclude.add("das_instance_id")
        include.add("das_endpoint") if (self.use_watsonx_credential_provider) else exclude.add("das_endpoint")
        (
            include.add("service_account_email")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.client_id)
                and (
                    (
                        (
                            self.authentication_method
                            and (
                                (
                                    hasattr(self.authentication_method, "value")
                                    and self.authentication_method.value == "workload_identity_federation_token"
                                )
                                or (self.authentication_method == "workload_identity_federation_token")
                            )
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "workload_identity_federation_token_url"
                            )
                            or (self.authentication_method == "workload_identity_federation_token_url")
                        )
                    )
                )
            )
            else exclude.add("service_account_email")
        )
        (
            include.add("service_account_token_lifetime")
            if (
                ((not self.credentials))
                and (not self.credentials_file_path)
                and (not self.client_id)
                and (
                    (
                        (
                            self.authentication_method
                            and (
                                (
                                    hasattr(self.authentication_method, "value")
                                    and self.authentication_method.value == "workload_identity_federation_token"
                                )
                                or (self.authentication_method == "workload_identity_federation_token")
                            )
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "workload_identity_federation_token_url"
                            )
                            or (self.authentication_method == "workload_identity_federation_token_url")
                        )
                    )
                )
            )
            else exclude.add("service_account_token_lifetime")
        )
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )
        include.add("credentials_file_path") if (self.hidden_dummy_property1) else exclude.add("credentials_file_path")
        include.add("vaulted_properties") if (self.hidden_dummy_property1) else exclude.add("vaulted_properties")
        include.add("ssl_certificate_file") if (self.hidden_dummy_property1) else exclude.add("ssl_certificate_file")

        (
            include.add("port_is_ssl_enabled")
            if (
                (self.proxy)
                and (
                    self.proxy_protocol
                    and (
                        (hasattr(self.proxy_protocol, "value") and self.proxy_protocol.value == "https")
                        or (self.proxy_protocol == "https")
                    )
                )
            )
            or (self.use_watsonx_credential_provider)
            else exclude.add("port_is_ssl_enabled")
        )
        include.add("proxy_port") if (self.proxy) else exclude.add("proxy_port")
        (
            include.add("ssl_certificate")
            if (not self.ssl_certificate_file) and (self.port_is_ssl_enabled == "true" or self.port_is_ssl_enabled)
            else exclude.add("ssl_certificate")
        )
        (
            include.add("service_account_email")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.client_id)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token_url" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token_url" in str(self.authentication_method))
                )
            )
            else exclude.add("service_account_email")
        )
        (
            include.add("token_type")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.client_id)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token_url" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token_url" in str(self.authentication_method))
                )
            )
            else exclude.add("token_type")
        )
        (
            include.add("request_body")
            if (
                self.http_method
                and (
                    (
                        hasattr(self.http_method, "value")
                        and self.http_method.value
                        and "post" in str(self.http_method.value)
                    )
                    or ("post" in str(self.http_method))
                )
                and self.http_method
                and (
                    (
                        hasattr(self.http_method, "value")
                        and self.http_method.value
                        and "put" in str(self.http_method.value)
                    )
                    or ("put" in str(self.http_method))
                )
            )
            else exclude.add("request_body")
        )
        include.add("das_api_key") if (self.use_watsonx_credential_provider) else exclude.add("das_api_key")
        (
            include.add("service_account_token_lifetime")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.client_id)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token_url" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token_url" in str(self.authentication_method))
                )
            )
            else exclude.add("service_account_token_lifetime")
        )
        include.add("proxy_username") if (self.proxy) else exclude.add("proxy_username")
        include.add("das_instance_id") if (self.use_watsonx_credential_provider) else exclude.add("das_instance_id")
        include.add("proxy_host") if (self.proxy) else exclude.add("proxy_host")
        (
            include.add("token_field_name")
            if (
                self.token_format
                and (
                    (hasattr(self.token_format, "value") and self.token_format.value == "json")
                    or (self.token_format == "json")
                )
            )
            else exclude.add("token_field_name")
        )
        include.add("proxy_protocol") if (self.proxy) else exclude.add("proxy_protocol")
        (
            include.add("credentials_file_path")
            if (not self.credentials)
            and (not self.client_id)
            and (not self.client_secret)
            and (not self.access_token)
            and (not self.refresh_token)
            and (not self.security_token_service_audience)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "credentials_file" in str(self.authentication_method.value)
                    )
                    or ("credentials_file" in str(self.authentication_method))
                )
            )
            and (not self.defer_credentials)
            and (not self.use_watsonx_credential_provider)
            else exclude.add("credentials_file_path")
        )
        (
            include.add("token_format")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.client_id)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token_url" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token_url" in str(self.authentication_method))
                )
            )
            else exclude.add("token_format")
        )
        (
            include.add("client_id")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.security_token_service_audience)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "credentials_oauth2" in str(self.authentication_method.value)
                    )
                    or ("credentials_oauth2" in str(self.authentication_method))
                )
            )
            and (not self.defer_credentials)
            and (not self.use_watsonx_credential_provider)
            else exclude.add("client_id")
        )
        (
            include.add("http_method")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.access_token)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value == "workload_identity_federation_token_url"
                    )
                    or (self.authentication_method == "workload_identity_federation_token_url")
                )
            )
            else exclude.add("http_method")
        )
        include.add("proxy_password") if (self.proxy) else exclude.add("proxy_password")
        (
            include.add("credentials")
            if (not self.credentials_file_path)
            and (not self.client_id)
            and (not self.client_secret)
            and (not self.access_token)
            and (not self.refresh_token)
            and (not self.security_token_service_audience)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "credentials" in str(self.authentication_method.value)
                    )
                    or ("credentials" in str(self.authentication_method))
                )
            )
            and (not self.defer_credentials)
            and (not self.use_watsonx_credential_provider)
            else exclude.add("credentials")
        )
        (
            include.add("client_secret")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.security_token_service_audience)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "credentials_oauth2" in str(self.authentication_method.value)
                    )
                    or ("credentials_oauth2" in str(self.authentication_method))
                )
            )
            and (not self.defer_credentials)
            and (not self.use_watsonx_credential_provider)
            else exclude.add("client_secret")
        )
        (
            include.add("validate_ssl_certificate")
            if (self.port_is_ssl_enabled == "true" or self.port_is_ssl_enabled)
            else exclude.add("validate_ssl_certificate")
        )
        (
            include.add("token_url")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.access_token)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value == "workload_identity_federation_token_url"
                    )
                    or (self.authentication_method == "workload_identity_federation_token_url")
                )
            )
            else exclude.add("token_url")
        )
        (
            include.add("http_headers")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.access_token)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value == "workload_identity_federation_token_url"
                    )
                    or (self.authentication_method == "workload_identity_federation_token_url")
                )
            )
            else exclude.add("http_headers")
        )
        include.add("das_endpoint") if (self.use_watsonx_credential_provider) else exclude.add("das_endpoint")
        (
            include.add("ssl_certificate_file")
            if (not self.ssl_certificate) and (self.port_is_ssl_enabled == "true" or self.port_is_ssl_enabled)
            else exclude.add("ssl_certificate_file")
        )
        (
            include.add("access_token")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.token_url)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "credentials_oauth2" in str(self.authentication_method.value)
                    )
                    or ("credentials_oauth2" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token" in str(self.authentication_method))
                )
            )
            and (not self.defer_credentials)
            and (not self.use_watsonx_credential_provider)
            else exclude.add("access_token")
        )
        (
            include.add("refresh_token")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.security_token_service_audience)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "credentials_oauth2" in str(self.authentication_method.value)
                    )
                    or ("credentials_oauth2" in str(self.authentication_method))
                )
            )
            and (not self.defer_credentials)
            and (not self.use_watsonx_credential_provider)
            else exclude.add("refresh_token")
        )
        (
            include.add("security_token_service_audience")
            if (not self.credentials)
            and (not self.credentials_file_path)
            and (not self.client_id)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "workload_identity_federation_token_url" in str(self.authentication_method.value)
                    )
                    or ("workload_identity_federation_token_url" in str(self.authentication_method))
                )
            )
            else exclude.add("security_token_service_audience")
        )
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)



import json
import re
from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.enums import FIELD
from pydantic import BaseModel, ConfigDict, Discriminator, Field, RootModel, Tag
from typing import Annotated, Any, Literal, Union


class Version(Enum):
    v3 = "3.0"


class JsonSchema(Enum):
    http = "http://api.dataplatform.ibm.com/schemas/common-pipeline/pipeline-flow/pipeline-flow-v3-schema.json"
    https = "https://api.dataplatform.ibm.com/schemas/common-pipeline/pipeline-flow/pipeline-flow-v3-schema.json"


class SubflowRef(BaseModel):
    url: str | None = None
    pipeline_id_ref: str
    name: str | None = None


class JsonSchema1(Enum):
    datarecord_schema = "https://api.dataplatform.ibm.com/schemas/common-pipeline/datarecord-metadata/datarecord-metadata-v3-schema.json"


class Measure(Enum):
    range = "range"
    discrete = "discrete"
    flag = "flag"
    set = "set"
    ordered_set = "ordered-set"
    typeless = "typeless"
    collection = "collection"
    geospatial = "geospatial"
    default = "default"

    def __str__(self):
        return self.name


class ModelingRole(Enum):
    input = "input"
    target = "target"
    both = "both"
    none = "none"
    partition = "partition"
    split = "split"
    frequency = "frequency"
    record_id = "record-id"

    def __str__(self):
        return self.name


class Range(BaseModel):
    min: str | float
    max: str | float


class Metadata(BaseModel):
    description: str | None = None
    measure: Measure | None = None
    modeling_role: ModelingRole | None = None
    max_length: int | None = None
    min_length: int | None = None
    decimal_precision: int | None = None
    decimal_scale: int | None = None
    values: list[str | int | bool] | None = None
    range: Range | None = None
    runtime_type: str | None = None
    is_key: bool | None = None
    is_signed: bool | None = None
    item_index: int | None = None
    source_field_id: str | None = None

    def __str__(self):
        dictionary = {
            "description": self.description,
            "measure": self.measure,
            "modeling_role": self.modeling_role,
            "max_length": self.max_length,
            "min_length": self.min_length,
            "decimal_precision": self.decimal_precision,
            "decimal_scale": self.decimal_scale,
            "values": self.values,
            "range": self.range,
            "runtime_type": self.runtime_type,
            "is_key": self.is_key,
            "is_signed": self.is_signed,
            "item_index": self.item_index,
            "source_field_id": self.source_field_id,
        }

        # remove all fields that is None.
        for key, value in dictionary.copy().items():
            if value is None:
                dictionary.pop(key)

        return json.dumps(dictionary, indent=4)


class AppDataDef(BaseModel):
    pass


class ParamSet(BaseModel):
    name: str | None = None
    ref: str
    catalog_ref: str | None = None
    project_ref: str | None = None
    space_ref: str | None = None


class PipelineFlowUI(BaseModel):
    model_config = ConfigDict(extra="allow")

    name: str | None = None
    description: str | None = None
    class_name: str | None = None


class ZoomObject(BaseModel):
    model_config = ConfigDict(extra="forbid")

    x: float
    y: float
    k: float


class ZoomObjectDef(RootModel[Union[int, ZoomObject]]):
    pass


class CommentLink(BaseModel):
    model_config = ConfigDict(extra="allow")

    node_ref: str
    class_name: str | None = None
    style: str | dict[str, Any] | None = None


class Type5(Enum):
    info = "info"
    error = "error"
    warning = "warning"


class Message(BaseModel):
    id_ref: str
    validation_id: str | None = None
    type: Type5
    text: str


class Position(Enum):
    topLeft = "topLeft"
    topCenter = "topCenter"
    topRight = "topRight"
    middleLeft = "middleLeft"
    middleCenter = "middleCenter"
    middleRight = "middleRight"
    bottomLeft = "bottomLeft"
    bottomCenter = "bottomCenter"
    bottomRight = "bottomRight"


class Position2(Enum):
    source = "source"
    middle = "middle"
    target = "target"


class LabelAlign(Enum):
    left = "left"
    center = "center"


class DecorationSharedProperties(BaseModel):
    id: str | None = None
    x_pos: float | None = None
    y_pos: float | None = None
    width: int | None = None
    height: int | None = None
    hotspot: bool | None = None
    class_name: str | None = None
    tooltip: str | None = None
    temporary: bool | None = None


class Cardinality(BaseModel):
    model_config = ConfigDict(extra="forbid")

    min: int | None = 1
    max: int | None = None


class PortUI(BaseModel):
    model_config = ConfigDict(extra="allow")

    cardinality: Cardinality | None = None
    class_name: str | None = None
    style: str | dict[str, Any] | None = None
    label: str | None = None


class RuntimeUI(BaseModel):
    model_config = ConfigDict(extra="allow")

    pass


class CommonPipelineConnection(BaseModel):
    name: str | None = None
    app_data: AppDataDef | None = None
    ref: str
    catalog_ref: str | None = None
    project_ref: str | None = None
    space_ref: str | None = None
    properties: dict[str, Any] | None = None
    connData: dict[str, Any] | None = None


class Properties(BaseModel):
    model_config = ConfigDict(extra="allow")

    attachment_ref: str | None = None
    name: str | None = False
    no_write_schema: bool | None = False
    no_write_status: bool | None = False


class CommonPipelineDataAsset(BaseModel):
    app_data: AppDataDef | None = None
    ref: str | None = None
    catalog_ref: str | None = None
    project_ref: str | None = None
    space_ref: str | None = None
    properties: Properties | None = None


class AppData(BaseModel):
    model_config = ConfigDict(extra="allow")

    ui_data: PipelineFlowUI | None = None


class PortAppData(BaseModel):
    model_config = ConfigDict(extra="allow")

    ui_data: PortUI | None = None


class RuntimeAppData(BaseModel):
    model_config = ConfigDict(extra="allow")

    ui_data: RuntimeUI | None = None


class Runtime(BaseModel):
    id: str
    name: str
    version: str | None = None
    app_data: RuntimeAppData | None = None


class FieldModel(BaseModel):
    """Serializable proxy for the FieldModelComplex class. Used in serializing fields for use in schema JSONs."""

    name: str
    type: str | None = "unknown"
    nullable: bool | None = False
    metadata: Metadata | None = None
    app_data: dict[str, Any] | None = None


class FieldModelComplex(BaseModel):
    model_config = ConfigDict(use_enum_values=True, populate_by_name=True)
    name: str
    type: str | None = "unknown"
    type_code: str | None = None
    odbc_type: str
    nullable: bool | None = False
    metadata: Metadata | None = None
    app_data: dict[str, Any] | None = None
    length: int | None = None
    scale: int | None = None
    signed: bool | None = Field(True)
    description: str | None = ""
    primary_key: bool | None = False
    unicode: FIELD.Unicode | None = Field(None, alias="extended")
    sqltype: int | None = 1
    ds_data_type: str | None = "WSMQ.QUEUENAME"
    data_element: str | None = "null"
    key_position: int | None = 1
    display_size: int | None = 0
    field_pos: int | None = 1
    level_no: int | None = None
    occurs: int | None = None
    sign_option: int | None = 0
    ds_scd_purpose: int | None = 0
    ds_sync_indicator: int | None = 0
    pad_char: str | None = ""  # TODO: Why are there two padchars (pad_char and padchar)
    extended_precision: int | None = 0
    tagged_subrec: int | None = 0
    occurs_varying: int | None = 0
    Maps_from_input_column: str | None = Field("", alias="Maps from input column")
    byte_to_skip: int | str | None = Field("", alias="skip")
    delim: FIELD.Delim | None = None
    delim_string: str | None = None
    generate: bool | None = False
    prefix_bytes: FIELD.Prefix | None = Field(None, alias="prefix")
    vector_prefix: FIELD.VectorPrefix | None = (
        None  # TODO: figure out why 0 is the default, but is not in the Prefix Enum
    )
    quote: FIELD.Quote | None = None
    start_position: int | None = Field(None, alias="position")
    tagcase: int | None = None
    charset: FIELD.CharSet | None = None
    default: int | str | None = None
    export_ebcdic_as_ascii: bool | None = Field(False)
    max_width: int | None = None
    width: int | None = max_width
    link_keep: bool | None = Field(False)
    padchar: FIELD.PadChar | None = None
    cycle_increment: FIELD.CycleIncrement | int | None = None
    cycle_initial_value: FIELD.CycleInitialValue | int | None = None
    cycle_limit: FIELD.CycleLimit | int | None = None
    cycle_values: list[str | int] | None = None
    alphabet: str | None = None
    time_scale: int | None = None
    byte_order: FIELD.ByteOrder | None = None
    c_format: str | None = Field(None)  # not sure how the handle the fact that this should only be included sometimes
    data_format: FIELD.DataFormat | None = None

    out_format: str | None = None
    reference: str | None = None
    generate_algorithm: FIELD.GenerateAlgorithm | None = None
    generate_type: FIELD.GenerateType | None = None
    random_limit: FIELD.RandomLimit | None = None
    random_seed: FIELD.RandomSeed | None = None
    random_signed: bool | None = False

    days_since: int | None = None
    date_format: str | None = None
    julian: bool | None = None

    generated_percent_invalid: int = Field(None, alias="invalids")
    epoch: int | None = None
    use_current_date: bool = Field(False)

    decimal_separator: FIELD.DecimalSeparator | None = None
    rounding: FIELD.Round | None = Field(None, alias="round")
    decimal_packed: FIELD.DecimalPacked | None = None
    check_decimal_packed: bool | None = Field(None, alias="check")
    decimal_packed_signed: bool | None = None
    sign_position: FIELD.SignPosition | None = None
    generated_percent_zeros: int = Field(None, alias="zeros")
    increment_scale: int | None = None
    precision: int | None = None

    midnight_seconds: bool = Field(False)
    time_scale_factor: int = Field(None, alias="_scale")
    extended_type: FIELD.TimeExtendedType | None = None
    allow_all_zeros: FIELD.AllowAllZeros | None = None

    timestamp_format: str | None = None
    vector_length_type: FIELD.VectorType | None = None
    apt_field_properties: str | None = None

    dimension_min_size: int | None = None

    actual_length: int | None = None
    null_length: int | None = None
    null_field: str | None = None
    nullseed: int | None = None
    nulls: int | None = None

    cycle: str | None = None

    def get_length(self):
        # return value, or return default derived from data type
        if self.length is not None:
            return self.length
        elif self.odbc_type in ["TIME"]:
            return 8
        elif self.odbc_type in ["TIMESTAMP"]:
            return 19
        elif self.odbc_type in [
            "SMALLINT",
            "INTEGER",
            "FLOAT",
            "DOUBLE",
            "DATE",
            "BIT",
            "TINYINT",
        ]:
            return 0
        else:
            return 100

    def get_scale(self):
        if self.scale is not None:
            return self.scale

        if self.extended_type is not None and self.extended_type != "" and self.extended_type != " ":
            return 6  # default value if extended type is set (see TIME and TIMESTAMP data types)

        return 0  # normal default value

    def get_cycle(self):
        if self.cycle:
            return self.cycle
        if self.generate_type == FIELD.GenerateType.cycle:
            cycle_dict_elements = []
            limit = self.cycle_limit
            if limit:
                limit = FieldModelComplex._get_enum_value(limit)
                if isinstance(limit, str):
                    cycle_dict_elements += [f"limit='{limit}'"]
                elif isinstance(limit, int):
                    cycle_dict_elements += [f"limit={limit}"]
            increment = self.cycle_increment
            if increment:
                increment = FieldModelComplex._get_enum_value(increment)
                if isinstance(increment, str):
                    cycle_dict_elements += [f"incr='{increment}'"]
                elif isinstance(increment, int):
                    cycle_dict_elements += [f"incr={increment}"]
            initial_value = self.cycle_initial_value
            if initial_value:
                initial_value = FieldModelComplex._get_enum_value(initial_value)
                if isinstance(initial_value, str):
                    cycle_dict_elements += [f"init='{initial_value}'"]
                elif isinstance(initial_value, int):
                    cycle_dict_elements += [f"init={initial_value}"]
            return "{" + ",".join(cycle_dict_elements) + "}"
        elif self.generate_algorithm == FIELD.GenerateAlgorithm.cycle:
            cycle_dict_elements = []
            cycle_values = self.cycle_values
            if cycle_values:
                for val in cycle_values:
                    if isinstance(val, str):
                        cycle_dict_elements += [f"value='{val}'"]
                    elif isinstance(val, int):
                        cycle_dict_elements += [f"value={val}"]
            return "{" + ",".join(cycle_dict_elements) + "}"
        return ""

    def get_random(self):
        random_dict_elements = []
        limit = self.random_limit
        if limit:
            limit = FieldModelComplex._get_enum_value(limit)
            if isinstance(limit, str):
                random_dict_elements += [f"limit='{limit}'"]
            elif isinstance(limit, int):
                random_dict_elements += [f"limit={limit}"]
        seed = self.random_seed
        if seed:
            seed = FieldModelComplex._get_enum_value(seed)
            if isinstance(seed, str):
                random_dict_elements += [f"seed='{seed}'"]
            elif isinstance(seed, int):
                random_dict_elements += [f"seed={seed}"]
        rand_signed = self.random_signed
        if rand_signed:
            random_dict_elements += ["signed"]
        return "{" + ",".join(random_dict_elements) + "}"

    @staticmethod
    def from_field_model(field_model: FieldModel):
        conversion_args = {
            "name": field_model.name,
            "type": field_model.type,
            "app_data": field_model.app_data,
            "metadata": field_model.metadata,
            "nullable": field_model.nullable,
        }
        if "odbc_type" in field_model.app_data:
            conversion_args["odbc_type"] = field_model.app_data["odbc_type"]
        else:
            raise Exception(
                "FieldModel was created without odbc data type, conversion to FieldModelComplex is impossible"
            )

        if "apt_field_properties" in field_model.app_data:
            conversion_args["apt_field_properties"] = field_model.app_data["apt_field_properties"]

        if (
            "dimension_min_size" in field_model.app_data
            and "dimension_max_size" in field_model.app_data
            and field_model.app_data["dimension_min_size"] == field_model.app_data["dimension_max_size"]
        ):
            conversion_args["occurs"] = field_model.app_data["dimension_min_size"]

        unicode = field_model.app_data.get("is_unicode_string", None)
        if unicode is not None:
            if unicode:
                conversion_args["unicode"] = FIELD.Unicode.true
            else:
                conversion_args["unicode"] = FIELD.Unicode.false

        extended_type = field_model.app_data.get("extended_type", None)
        if extended_type is not None:
            conversion_args["extended_type"] = FIELD.TimeExtendedType(extended_type)

        # if length is a consistent value (always if not default), then set length property. Otherwise rely on existing default setting logic
        if (
            (
                conversion_args["odbc_type"] in ["BINARY", "CHAR", "WCHAR"]
                and field_model.metadata.min_length == field_model.metadata.max_length
            )
            or (
                conversion_args["odbc_type"] in ["DECIMAL", "NUMERIC", "TIME", "TIMESTAMP"]
                and field_model.metadata.decimal_precision == field_model.metadata.max_length
            )
            or (
                conversion_args["odbc_type"]
                in [
                    "WLONGVARCHAR",
                    "LONGVARBINARY",
                    "LONGVARCHAR",
                    "WVARCHAR",
                    "REAL",
                    "UNKNOWN",
                    "VARBINARY",
                    "VARCHAR",
                ]
            )
        ):
            conversion_args["length"] = field_model.metadata.max_length

        conversion_args["time_scale"] = field_model.app_data.get("time_scale", 0)

        conversion_args["scale"] = field_model.metadata.decimal_scale

        conversion_args["primary_key"] = field_model.metadata.is_key
        conversion_args["description"] = field_model.metadata.description
        conversion_args["signed"] = field_model.metadata.is_signed
        conversion_args["level_no"] = (
            field_model.metadata.item_index if field_model.metadata.item_index is not None else 0
        )

        return FieldModelComplex(**conversion_args)

    def get_field_model(self):
        self.validate_properties()
        extended_properties_app_data = self.app_data.copy()

        extended_properties_app_data["odbc_type"] = self.odbc_type

        apt_props = self.get_aptfieldproperties()
        if apt_props != "":
            extended_properties_app_data["apt_field_properties"] = apt_props

        if self.occurs is not None:
            extended_properties_app_data["dimension_min_size"] = f"{self.occurs}"
            extended_properties_app_data["dimension_max_size"] = f"{self.occurs}"

        if (
            self.vector_length_type == FIELD.VectorType.variable
            or self.vector_length_type == FIELD.VectorType.variable.value
        ):
            extended_properties_app_data["dimension_min_size"] = "0"

        if self.unicode == FIELD.Unicode.true or self.unicode == FIELD.Unicode.true.value:
            extended_properties_app_data["is_unicode_string"] = True
        elif self.unicode == FIELD.Unicode.false or self.unicode == FIELD.Unicode.false.value:
            extended_properties_app_data["is_unicode_string"] = False

        if self.extended_type is not None:
            extended_properties_app_data["extended_type"] = (
                self.extended_type.value
                if isinstance(self.extended_type, FIELD.TimeExtendedType)
                else self.extended_type
            )

        if self.length is not None:
            if self.odbc_type in ["BINARY", "CHAR", "WCHAR"]:
                self.metadata.min_length = self.length
                self.metadata.max_length = self.length
            elif self.odbc_type in ["DECIMAL", "NUMERIC", "TIME", "TIMESTAMP"]:
                self.metadata.max_length = self.length
                self.metadata.decimal_precision = self.length
            elif self.odbc_type in [
                "WLONGVARCHAR",
                "LONGVARBINARY",
                "LONGVARCHAR",
                "WVARCHAR",
                "REAL",
                "UNKNOWN",
                "VARBINARY",
                "VARCHAR",
            ]:
                self.metadata.max_length = self.length

        self.metadata.decimal_scale = self.get_scale()

        if self.time_scale is not None:
            self.app_data["time_scale"] = self.time_scale

        self.metadata.is_key = self.primary_key
        self.metadata.description = self.description
        self.metadata.is_signed = self.signed
        self.metadata.item_index = self.level_no if self.level_no is not None else 0

        if self.type_code is not None:
            extended_properties_app_data["type_code"] = self.type_code

        return FieldModel(
            name=self.name,
            type=self.type,
            nullable=self.nullable,
            metadata=self.metadata,
            app_data=extended_properties_app_data,
        )

    def serialize_to_data_definition_column(self):
        self.validate_properties()
        dumped = {"name": self.name}
        dumped["type"] = self.get_type_props()
        dumped["properties"] = self.get_properties_props()

        return dumped

    def get_type_props(self):
        props = {"length", "scale", "nullable", "signed"}
        dumped = self.model_dump(include=props, by_alias=True)
        dumped["length"] = self.get_length()
        dumped["scale"] = self.get_scale()
        type_mapping = {
            "WCHAR": "NCHAR",
            "WVARCHAR": "NVARCHAR",
            "WLONGVARCHAR": "LONGNVARCHAR",
        }
        dumped["type"] = type_mapping.get(self.odbc_type, self.odbc_type)

        return dumped

    def get_properties_props(self):
        include_props: set[str] = {
            "Maps_from_input_column",
            "alphabet",
            "charset",
            "data_element",
            "default",
            "delim",
            "delim_string",
            "description",
            "display_size",
            "ds_data_type",
            "ds_scd_purpose",
            "ds_sync_indicator",
            "export_ebcdic_as_ascii",
            "unicode",
            "extended_precision",
            "field_pos",
            "generate",
            "key_position",
            "link_keep",
            "max_width",
            "occurs_varying",
            "pad_char",
            "padchar",
            "start_position",
            "primary_key",
            "quote",
            "sign_option",
            "byte_to_skip",
            "sqltype",
            "tagcase",
            "tagged_subrec",
            "time_scale",
            "width",
        }
        # add optional properties
        optional_props = {"occurs", "level_no", "extended_type"}
        include_props.update(optional_props)

        # special case for TIMESTAMP data type
        if self.extended_type is None and self.odbc_type == "TIMESTAMP":
            self.extended_type = " "

        dumped: dict = json.loads(
            self.model_dump_json(
                include=include_props,
                by_alias=True,
            )
        )

        # remove optional properties if unset
        for optional_prop in optional_props:
            if dumped[optional_prop] is None:
                dumped.pop(optional_prop)

        dumped["APTFieldProperties"] = self.get_aptfieldproperties()
        dumped["MisFieldProperties"] = self.get_misfieldproperties()
        default_overrides_for_unset_apt_properties = {
            "position": 0,
            "tagcase": 0,
            "default": "",
            "max_width": 0,
            "width": 0,
            "charset": "",
            "padchar": "",
            "delim": "",
            "delim_string": "",
            "time_scale": 0,
            "quote": "",
            "alphabet": "",
        }

        for key, value in default_overrides_for_unset_apt_properties.items():
            if dumped[key] is None:
                dumped[key] = value

        # dealing with dominance of vector prefix over prefix
        if self.vector_prefix is not None:
            dumped["prefix"] = FieldModelComplex._get_enum_value(self.vector_prefix)
        elif self.prefix_bytes is not None:
            dumped["prefix"] = FieldModelComplex._get_enum_value(self.prefix_bytes)
        else:
            dumped["prefix"] = 0

        dumped["cycle"] = self.get_cycle()

        return dumped

    def get_aptfieldproperties(self):
        if self.apt_field_properties is not None:
            return self.apt_field_properties
        string_props: set[str] = {
            "c_format",
            "default",
            "out_format",
            "padchar",
            "reference",
            "date_format",
            "decimal_separator",
            "timestamp_format",
            "delim_string",
            "delim",
            "null_field",
        }
        integer_props: set[str] = {
            "start_position",
            "tagcase",
            "default",
            "max_width",
            "width",
            "prefix_bytes",
            "vector_prefix",
            "days_since",
            "generated_percent_invalid",
            "epoch",
            "generated_percent_zeros",
            "precision",
            "time_scale_factor",
            "byte_to_skip",
            "actual_length",
            "null_length",
            "nullseed",
            "nulls",
        }
        # not sure if default could ever be a boolean
        boolean_props: set[str] = {
            "generate",
            "link_keep",
            "export_ebcdic_as_ascii",
            "julian",
            "check_decimal_packed",
            "midnight_seconds",
        }
        # Some enums may be included in strings if they are represented internally that way
        enum_props: set[str] = {
            "byte_order",
            "charset",
            "data_format",
            "sign_position",
            "allow_all_zeros",
            "decimal_packed",
        }
        all_props: set[str] = string_props | integer_props | boolean_props | enum_props
        prop_dict: dict = json.loads(self.model_dump_json(include=all_props, by_alias=True))
        fields: list[str] = []
        integer_prop_aliases = {
            "position",
            "prefix",
            "invalids",
            "zeros",
            "_scale",
            "skip",
        }
        boolean_prop_aliases = {"generate", "check"}
        enum_prop_aliases = {}
        for key, value in prop_dict.items():
            if key == "padchar" and value == "null":
                fields += [f"{key}={value}"]
            elif key == "round" and value:
                fields += [f"{key}={value}"]
            elif (key in string_props) and isinstance(value, str):
                fields += [f"{key}='{value}'"]
            elif (key in integer_props or key in integer_prop_aliases) and isinstance(value, int):
                fields += [f"{key}={value}"]
            elif (key in boolean_props or key in boolean_prop_aliases) and isinstance(value, bool) and value:
                fields += [f"{key}"]
            elif (key in enum_props or key in enum_prop_aliases) and value is not None:
                fields += [f"{value}"]

        cycle_val = self.get_cycle()
        if cycle_val:
            fields += [f"cycle={cycle_val}"]

        if self.generate_type == FIELD.GenerateType.random:
            fields += [f"random={self.get_random()}"]
        elif self.generate_algorithm == FIELD.GenerateAlgorithm.alphabet:
            fields += [f"alphabet={self.alphabet}"]

        if self.use_current_date:
            fields += ["function=rundate"]

        if self.rounding:
            fields += [f"round={self.rounding.value}"]

        if self.increment_scale is not None:
            fields += [f"scale={self.increment_scale}"]

        if self.decimal_packed_signed:
            fields += ["signed"]

        if self.quote is not None:
            fields += [f"quote={self.quote.value}"]

        return ",".join(fields)

    def get_misfieldproperties(self):
        fields: list[str] = []
        if self.occurs is not None:
            fields += [f"dimension_min_size={self.occurs}"]
        elif self.dimension_min_size is not None:
            fields += [f"dimension_min_size={self.dimension_min_size}"]
        if self.level_no is not None:
            fields += [f"LevelNumber={self.level_no}"]
        return "|".join(fields)

    @classmethod
    def from_data_definition_column_dict(cls, _dict: dict) -> "FieldModelComplex":
        """Initialize a DataDefinitionColumn object from a json dictionary."""
        args = {}
        if (name := _dict.get("name")) is not None:
            args["name"] = name
        else:
            raise ValueError("Required property 'name' not present in DataDefinitionColumn JSON")
        if (type := _dict.get("type").get("type")) is not None:
            args["odbc_type"] = type
        else:
            raise ValueError("Required property 'type' not present in DataDefinitionColumn JSON")
        if (length := _dict.get("type").get("length")) is not None:
            args["length"] = length
        else:
            args["length"] = None
        if (scale := _dict.get("type").get("scale")) is not None:
            args["scale"] = scale
        else:
            args["scale"] = None
        if (nullable := _dict.get("type").get("nullable")) is not None:
            args["nullable"] = nullable
        else:
            args["nullable"] = None
        if (signed := _dict.get("type").get("signed")) is not None:
            args["signed"] = signed
        else:
            args["signed"] = None
        properties_keys = {
            "APTFieldProperties",
            "Maps from input column",
            # "MisFieldProperties",
            "alphabet",
            "charset",
            "cycle",
            "data_element",
            "default",
            "delim",
            "delim_string",
            "description",
            "display_size",
            "ds_data_type",
            "ds_scd_purpose",
            "ds_sync_indicator",
            "export_ebcdic_as_ascii",
            "extended",
            "extended_precision",
            "field_pos",
            "generate",
            "key_position",
            "link_keep",
            "max_width",
            "occurs_varying",
            "pad_char",
            "padchar",
            "position",
            "prefix",
            "primary_key",
            "quote",
            "sign_option",
            "skip",
            "sqltype",
            "tagcase",
            "tagged_subrec",
            "time_scale",
            "width",
        }

        optional_properties_keys = {"occurs"}

        enum_mappings = {
            "extended": FIELD.Unicode,
            "delim": FIELD.Delim,
            "quote": FIELD.Quote,
            "charset": FIELD.CharSet,
            "padchar": FIELD.PadChar,
        }

        inverse_alias_map = {
            "extended": "unicode",
            "level_no": "level_no",
            "Maps from input column": "Maps_from_input_column",
            "skip": "byte_to_skip",
            "generate": "generate",
        }
        # TODO: deal with prefix vs vector_prefix here, dominance, etc.

        properties_dict = _dict.get("properties")

        enums_with_invalid_empty_str_defaults = {
            "quote",
            "charset",
            "padchar",
            "delim",
            "delim_string",
        }
        for key in enums_with_invalid_empty_str_defaults:
            if properties_dict.get(key) == "":
                properties_dict[key] = None
                args[key] = None
                properties_keys.remove(key)

        apt_field_props_str = properties_dict.get("APTFieldProperties")
        if apt_field_props_str and "vector_prefix" in apt_field_props_str and properties_dict.get("prefix") != 0:
            args["vector_prefix"] = FIELD.VectorPrefix(properties_dict.get("prefix"))
        elif properties_dict.get("prefix", 0) != 0:
            args["prefix_bytes"] = FIELD.Prefix(properties_dict.get("prefix"))
        properties_keys.remove("prefix")
        args["prefix_bytes"] = args.get("prefix_bytes", None)

        if apt_field_props_str and ("vector_prefix" in apt_field_props_str or "reference" in apt_field_props_str):
            args["dimension_min_size"] = "0"

        args["vector_prefix"] = args.get("vector_prefix", None)

        for key in properties_keys:
            if (key_value := properties_dict.get(key)) is not None:
                args[inverse_alias_map.get(key, key)] = FieldModelComplex._get_value_from_json_value(
                    key_value, enum_mappings.get(key, None)
                )
            else:
                args[inverse_alias_map.get(key, key)] = None

        for key in optional_properties_keys:
            if (key_value := properties_dict.get(key)) is not None:
                args[inverse_alias_map.get(key, key)] = FieldModelComplex._get_value_from_json_value(
                    key_value, enum_mappings.get(key, None)
                )

        # parse misfieldproperties
        misfield_props = properties_dict.get("MisFieldProperties")
        if misfield_props:
            dim_min_size_search = re.search(r"dimension_min_size=(\d+)", misfield_props)
        else:
            dim_min_size_search = None
        if dim_min_size_search is not None:
            dim_min_size_value = int(dim_min_size_search.group(1))
            # if vector length is variable, then dim_min_size is set to 0 when vector occurs propogates
            if dim_min_size_value == 0 and properties_dict.get("occurs") != 0:
                args["vector_length_type"] = FIELD.VectorType.variable
                args["occurs"] = None
                args["dimension_min_size"] = "0"
            else:
                args["occurs"] = dim_min_size_value
        elif properties_dict.get("occurs", 0) != 0:
            args["occurs"] = properties_dict.get("occurs")
        else:
            args["occurs"] = None
        args["dimension_min_size"] = args.get("dimension_min_size", args["occurs"])

        args["vector_length_type"] = args.get("vector_length_type", None)

        if misfield_props:
            level_number_search = re.search(r"LevelNumber=(\d+)", misfield_props)
        else:
            level_number_search = None
        if level_number_search is not None:
            args["level_no"] = int(level_number_search.group(1))
        else:
            args["level_no"] = None

        odbc_type: str = args["odbc_type"]

        odbc_type_to_type = {
            "BIGINT": "integer",
            "BINARY": "binary",
            "BIT": "boolean",
            "CHAR": "string",
            "DATE": "date",
            "DECIMAL": "double",
            "DOUBLE": "double",
            "FLOAT": "double",
            "INTEGER": "integer",
            "WLONGVARCHAR": "string",
            "LONGNVARCHAR": "string",
            "LONGVARBINARY": "binary",
            "LONGVARCHAR": "string",
            "WCHAR": "string",
            "NCHAR": "string",
            "NUMERIC": "double",
            "WVARCHAR": "string",
            "NVARCHAR": "string",
            "REAL": "double",
            "SMALLINT": "integer",
            "TIME": "time",
            "TIMESTAMP": "timestamp",
            "TINYINT": "integer",
            "UNKNOWN": "string",
            "VARBINARY": "binary",
            "VARCHAR": "string",
        }
        args["type"] = odbc_type_to_type[odbc_type]

        odbc_type_to_type_code = {
            "BIGINT": "INT64",
            "BINARY": "BINARY",
            "BIT": "BOOLEAN",
            "CHAR": "STRING",
            "DATE": "DATE",
            "DECIMAL": "DECIMAL",
            "DOUBLE": "DFLOAT",
            "FLOAT": "SFLOAT",
            "INTEGER": "INT32",
            "WLONGVARCHAR": "STRING",
            "LONGNVARCHAR": "STRING",
            "LONGVARBINARY": "BINARY",
            "LONGVARCHAR": "STRING",
            "WCHAR": "STRING",
            "NCHAR": "STRING",
            "NUMERIC": "DECIMAL",
            "WVARCHAR": "STRING",
            "NVARCHAR": "STRING",
            "REAL": "SFLOAT",
            "SMALLINT": "INT16",
            "TIME": "TIME",
            "TIMESTAMP": "DATETIME",
            "TINYINT": "INT8",
            "UNKNOWN": "UNKNOWN",
            "VARBINARY": "BINARY",
            "VARCHAR": "STRING",
        }

        args["type_code"] = odbc_type_to_type_code[odbc_type]

        imported_metadata = Metadata(
            is_key=False,
            item_index=0,
            is_signed=True,
            description="",
            min_length=0,
            decimal_precision=100,
            decimal_scale=0,
        )
        # defaults: (min_length, max_length, decimal_precision, decimal_scale)
        odbc_type_to_metadata_updates = {
            "BIGINT": (0, 6),
            "BINARY": (100, 100, 0),
            "BIT": (6, 6),
            "CHAR": (100, 100, 0),
            "DATE": (0, 10),
            "DECIMAL": (0, 100),
            "DOUBLE": (0, 6),
            "FLOAT": (0, 6),
            "INTEGER": (0, 6),
            "WLONGVARCHAR": (0, 100),
            "LONGNVARCHAR": (0, 100),
            "LONGVARBINARY": (0, 100),
            "LONGVARCHAR": (0, 100),
            "WCHAR": (100, 100),
            "NCHAR": (100, 100),
            "NUMERIC": (0, 100),
            "WVARCHAR": (0, 100),
            "NVARCHAR": (0, 100),
            "REAL": (0, 100),
            "SMALLINT": (0, 6),
            "TIME": (0, 8, 8, 0),
            "TIMESTAMP": (0, 19, 19, 0),
            "TINYINT": (0, 6),
            "UNKNOWN": (0, 100),
            "VARBINARY": (0, 100),
            "VARCHAR": (0, 100),
        }

        metadata_updates = odbc_type_to_metadata_updates[odbc_type]

        imported_metadata.min_length = metadata_updates[0]
        imported_metadata.max_length = metadata_updates[1]
        if len(metadata_updates) > 2:
            imported_metadata.decimal_precision = metadata_updates[2]
        if len(metadata_updates) > 3:
            imported_metadata.decimal_scale = metadata_updates[3]

        args["metadata"] = imported_metadata
        args["app_data"] = {"time_scale": args.get("time_scale", 0)}

        if "extended_type" in properties_dict and properties_dict["extended_type"] != " ":
            args["extended_type"] = FieldModelComplex._get_value_from_json_value(
                properties_dict["extended_type"], FIELD.TimeExtendedType
            )
        args["extended_type"] = args.get("extended_type", None)

        return FieldModelComplex(
            name=args["name"],
            metadata=args["metadata"],
            app_data=args["app_data"],
            type=args["type"],
            type_code=args["type_code"],
            odbc_type=args["odbc_type"],
            length=args["length"],
            scale=args["scale"],
            nullable=args["nullable"],
            signed=args["signed"],
            pad_char=args["pad_char"],
            primary_key=args["primary_key"],
            max_width=args["max_width"],
            tagged_subrec=args["tagged_subrec"],
            width=args["width"],
            Maps_from_input_column=args["Maps_from_input_column"],
            quote=args["quote"],
            tagcase=args["tagcase"],
            alphabet=args["alphabet"],
            delim_string=args["delim_string"],
            cycle=args["cycle"],
            byte_to_skip=args["byte_to_skip"],
            generate=args["generate"],
            ds_sync_indicator=args["ds_sync_indicator"],
            ds_scd_purpose=args["ds_scd_purpose"],
            vector_prefix=args["vector_prefix"],
            prefix_bytes=args["prefix_bytes"],
            display_size=args["display_size"],
            padchar=args["padchar"],
            data_element=args["data_element"],
            field_pos=args["field_pos"],
            description=args["description"],
            position=args["position"],
            sign_option=args["sign_option"],
            default=args["default"],
            sqltype=args["sqltype"],
            occurs_varying=args["occurs_varying"],
            time_scale=args["time_scale"],
            charset=args["charset"],
            unicode=args["unicode"],
            delim=args["delim"],
            key_position=args["key_position"],
            ds_data_type=args["ds_data_type"],
            apt_field_properties=args["APTFieldProperties"],
            extended_precision=args["extended_precision"],
            link_keep=args["link_keep"],
            export_ebcdic_as_ascii=args["export_ebcdic_as_ascii"],
            occurs=args["occurs"],
            level_no=args["level_no"],
            vector_length_type=args["vector_length_type"],
            dimension_min_size=args["dimension_min_size"],
            extended_type=args["extended_type"],
        )

    def _get_value_from_json_value(internal_value: str | int, enum_type: Enum | None):
        if enum_type is None:
            return internal_value
        return enum_type(internal_value)

    def _get_enum_value(possible_enum: Enum | int | str):
        if isinstance(possible_enum, Enum):
            return possible_enum.value
        else:
            return possible_enum

    def validate_properties(self) -> None:
        prop_to_incompatible_props: dict[str, set] = {
            "date_format": {"days_since", "julian", "data_format"},
            "reference": {"delim_string", "delim", "quote", "prefix_bytes"},
            "prefix_bytes": {"delim_string", "delim", "quote", "reference"},
            "days_since": {"data_format", "julian", "date_format"},
            "julian": {"days_since", "data_format", "date_format"},
            "data_format": {"days_since", "julian", "date_format"},
            "delim": {"delim_string", "prefix_bytes"},
            "delim_string": {"delim", "prefix_bytes"},
            "quote": {"prefix_bytes"},
            "actual_length": {"null_field"},
            "null_length": {"null_field"},
            "null_field": {"actual_length", "null_length"},
        }
        for prop, incompatible_props in prop_to_incompatible_props.items():
            self.validate_prop_with_incompatibilities(prop, incompatible_props)

        sub_properties_to_valid_super_props: dict[str, list[tuple[str, Any]]] = {
            # subproperty: [(superproperty name, valid superproperty value), ...]
            "reference": [("vector_length_type", FIELD.VectorType.variable)],
            "vector_prefix": [("vector_length_type", FIELD.VectorType.variable)],
            "check_decimal_packed": [("decimal_packed", FIELD.DecimalPacked.packed)],
            "decimal_packed_signed": [
                ("decimal_packed", FIELD.DecimalPacked.packed),
                ("decimal_packed", FIELD.DecimalPacked.zoned),
            ],
            "sign_position": [
                ("decimal_packed", FIELD.DecimalPacked.overpunch),
                ("decimal_packed", FIELD.DecimalPacked.separate),
                ("decimal_packed", FIELD.DecimalPacked.zoned),
            ],
            "occurs": [("vector_length_type", FIELD.VectorType.vector_occurs)],
            "cycle_increment": [("generate_type", FIELD.GenerateType.cycle)],
            "cycle_initial_value": [("generate_type", FIELD.GenerateType.cycle)],
            "cycle_limit": [("generate_type", FIELD.GenerateType.cycle)],
            "cycle_values": [("generate_algorithm", FIELD.GenerateAlgorithm.cycle)],
            "alphabet": [("generate_algorithm", FIELD.GenerateAlgorithm.alphabet)],
            "actual_length": [("nullable", True)],
            "null_length": [("nullable", True)],
            "null_field": [("nullable", True)],
            "nullseed": [("nullable", True)],
            "nulls": [("nullable", True)],
        }

        for sub_prop, valid_super_props in sub_properties_to_valid_super_props.items():
            self.validate_subproperty(sub_prop, valid_super_props)

    def validate_prop_with_incompatibilities(self, prop: str, incompatible_props: set[str]):
        dumped = self.model_dump(include=(incompatible_props | {prop}))
        if dumped[prop] is not None:
            for incompat_prop in incompatible_props:
                if dumped[incompat_prop] is not None:
                    print(
                        f"Property '{prop}' set on field '{self.name}', despite incompatible property '{incompat_prop}' also being set."
                    )

    def validate_subproperty(self, sub_prop: str, valid_super_props: list[tuple[str, Any]]):
        super_prop_names = set([super_prop[0] for super_prop in valid_super_props])
        dumped = self.model_dump(include=(super_prop_names | {sub_prop}))
        if dumped[sub_prop] is None:
            return
        for super_prop_name, super_prop_valid_value in valid_super_props:
            if dumped[super_prop_name] == super_prop_valid_value:
                return
        print(
            f"Subproperty '{sub_prop}' set without required/compatible superproperties being set on field '{self.name}'. Possible superproperties that could be set for subproperty include: {super_prop_names}."
        )


class Comment(BaseModel):
    model_config = ConfigDict(extra="allow", populate_by_name=True)

    id: str
    x_pos: float
    y_pos: float
    width: Annotated[float, Field(ge=10.0)]
    height: Annotated[float, Field(ge=10.0)]
    class_name: str | None = None
    style: str | dict[str, Any] | None = None
    attributes: str | None = None
    content: str | None = None
    content_type: str | None = Field(None, alias="contentType")
    associated_id_refs: list[CommentLink] | None = None


class ImageDecoration(DecorationSharedProperties):
    image: str | None = None
    outline: bool | None = None


class LabelDecoration(DecorationSharedProperties):
    label: str | None = None
    label_editable: bool | None = None
    label_align: LabelAlign | None = None
    label_single_line: bool | None = None
    label_max_characters: int | None = None
    label_allow_return_key: bool | None = None


class ShapeDecoration(DecorationSharedProperties):
    path: str | None = None


class JsxDecoration(DecorationSharedProperties):
    jsx: dict[str, Any] | None = None


class StructTypes(BaseModel):
    fields: list[FieldModel] | None = None


class RecordSchema(BaseModel):
    id: str
    name: str | None = None
    json_schema: JsonSchema1 | None = None
    type: str | None = None
    fields: list[FieldModel | FieldModelComplex]
    struct_types: dict[Annotated[str, Field(pattern=r".")], StructTypes] | None = None


class PipelineUI(BaseModel):
    model_config = ConfigDict(extra="allow")

    description: str | None = None
    zoom: ZoomObjectDef | None = None
    comments: list[Comment] | None = None


class NodeDecoration1(ImageDecoration):
    position: Position | None = None


class NodeDecoration2(LabelDecoration):
    position: Position | None = None


class NodeDecoration3(ShapeDecoration):
    position: Position | None = None


class NodeDecoration4(JsxDecoration):
    position: Position | None = None


class NodeDecoration(RootModel[Union[NodeDecoration1, NodeDecoration2, NodeDecoration3, NodeDecoration4]]):
    pass


class LinkDecoration1(ImageDecoration):
    position: Position2 | None = None
    distance: int | None = None


class LinkDecoration2(LabelDecoration):
    position: Position2 | None = None
    distance: int | None = None


class LinkDecoration3(ShapeDecoration):
    position: Position2 | None = None
    distance: int | None = None


class LinkDecoration4(JsxDecoration):
    position: Position2 | None = None
    distance: int | None = None


class LinkDecoration(RootModel[Union[LinkDecoration1, LinkDecoration2, LinkDecoration3, LinkDecoration4]]):
    pass


class PipelineAppData(BaseModel):
    model_config = ConfigDict(extra="allow")

    ui_data: PipelineUI | None = None


class Decoration(RootModel[Union[NodeDecoration, LinkDecoration]]):
    pass


class NodeLinkUI(BaseModel):
    model_config = ConfigDict(extra="allow")

    description: str | None = None
    class_name: str | None = None
    style: str | dict[str, Any] | None = None
    decorations: list[Decoration] | None = None


class LinkAppData(BaseModel):
    model_config = ConfigDict(extra="allow")

    ui_data: NodeLinkUI | None = None


class NodeLink(BaseModel):
    model_config = ConfigDict(extra="forbid")

    id: str | None = None
    node_id_ref: str
    port_id_ref: str | None = None
    link_name: str | None = None
    type_attr: str | None = None
    description: str | None = None
    app_data: LinkAppData | None = None


class AssociationLink(BaseModel):
    model_config = ConfigDict(extra="forbid")

    id: str
    node_ref: str
    class_name: str | None = None
    style: str | dict[str, Any] | None = None
    decorations: list[Decoration] | None = None


class Port(BaseModel):
    model_config = ConfigDict(extra="forbid")

    id: str
    schema_ref: str | None = None
    links: list[NodeLink] | None = None
    parameters: dict[str, Any] | None = None
    app_data: PortAppData | None = None


class BoundPort(BaseModel):
    model_config = ConfigDict(extra="forbid")

    id: str = Field(...)
    schema_ref: str | None = None
    links: list[NodeLink] | None = None
    subflow_node_ref: str | None = None
    parameters: dict[str, Any] | None = None
    app_data: PortAppData | None = None


class NodeUI(BaseModel):
    model_config = ConfigDict(extra="allow")

    label: str | None = None
    description: str | None = None
    class_name: str | None = None
    style: str | dict[str, Any] | None = None
    image: str | dict[str, Any] | None = None
    x_pos: float | None = None
    y_pos: float | None = None
    is_expanded: bool | None = None
    expanded_height: float | None = None
    expanded_width: float | None = None
    is_resized: bool | None = None
    resize_height: float | None = None
    resize_width: float | None = None
    sub_pipelines: list | None = None
    attributes: str | None = None
    associations: list[AssociationLink] | None = None
    messages: list[Message] | None = None
    ui_parameters: dict[str, Any] | None = None
    palette_image: str | None = None
    palette_class_name: str | None = None
    palette_disabled: bool | None = None
    decorations: list[Decoration] | None = None


class NodeAppData(BaseModel):
    model_config = ConfigDict(extra="allow")

    ui_data: NodeUI | None = None


def get_node_discriminator(v: Any):
    if isinstance(v, dict):
        node_type = v.get("type")
        has_inputs = bool(v.get("inputs"))
    else:
        node_type = getattr(v, "type")
        has_inputs = hasattr(v, "inputs")

    if node_type == "binding":
        return "binding_exit_node" if has_inputs else "binding_entry_node"
    else:
        return node_type


class Node(BaseModel):
    model_config = ConfigDict(extra="forbid")

    type: str
    id: str
    description: str | None = None
    app_data: NodeAppData | None = None


class ExecutionNode(Node):
    type: Literal["execution_node"] = "execution_node"
    op: str
    inputs: list[Port] | None = None
    outputs: list[Port] | None = None
    parameters: dict[str, Any] | None = None
    runtime_ref: str | None = None


class BindingEntryNode(Node):
    type: Literal["binding"] = "binding"
    outputs: list[Port]
    connection: CommonPipelineConnection | None = None
    data_asset: CommonPipelineDataAsset | None = None
    op: str | None = None
    parameters: dict[str, Any] | None = None


class BindingExitNode(Node):
    type: Literal["binding"] = "binding"
    inputs: list[Port]
    outputs: list[Port] | None = None
    connection: CommonPipelineConnection | None = None
    data_asset: CommonPipelineDataAsset | None = None
    op: str | None = None
    parameters: dict[str, Any] | None = None


class ModelNode(Node):
    type: Literal["model_node"] = "model_node"
    model_ref: str | None = None
    inputs: list[Port]
    outputs: list[Port] | None = None
    op: str | None = None
    parameters: dict[str, Any] | None = None
    runtime_ref: str | None = None


class Supernode(Node):
    type: Literal["super_node"] = "super_node"
    open_with_tool: str | None = None
    subflow_ref: SubflowRef
    inputs: list[BoundPort] | None = None
    outputs: list[BoundPort] | None = None
    parameters: dict[str, Any] | None = None


class Pipeline(BaseModel):
    model_config = ConfigDict(extra="forbid")

    id: str
    description: str | None = None
    name: str | None = None
    runtime_ref: str
    nodes: list[
        Annotated[
            Annotated[Supernode, Tag("super_node")]
            | Annotated[BindingEntryNode, Tag("binding_entry_node")]
            | Annotated[BindingExitNode, Tag("binding_exit_node")]
            | Annotated[ModelNode, Tag("model_node")]
            | Annotated[ExecutionNode, Tag("execution_node")],
            Discriminator(get_node_discriminator),
        ]
    ]
    parameters: dict[str, Any] | None = None
    app_data: PipelineAppData | None = None


class Flow(BaseModel):
    model_config = ConfigDict(extra="forbid", use_enum_values=True)

    doc_type: str | None = "pipeline"
    version: Version = Version.v3
    json_schema: JsonSchema | None = JsonSchema.https
    open_with_tool: str | None = None
    id: str | None = None
    parameters: dict[str, Any] | None = None
    primary_pipeline: str
    pipelines: list[Pipeline]
    schemas: list[RecordSchema] | None = None
    runtimes: list[Runtime] | None = None
    external_paramsets: list[ParamSet] | None = None
    app_data: AppData | None = None
    name: str | None = None



# IBM Confidential
# PID 5900-BAF
# Copyright StreamSets Inc., an IBM Company 2025

"""This module contains models and abstractions for the StreamSets CPD Service."""

from ibm_watsonx_data_integration.services.streamsets.models.engine_model import Engine
from ibm_watsonx_data_integration.services.streamsets.models.environment_model import Environment, Environments
from ibm_watsonx_data_integration.services.streamsets.models.flow_model import (
    StreamsetsConnection,
    StreamsetsFlow,
    StreamsetsFlows,
)

__all__ = [
    "StreamsetsFlow",
    "StreamsetsFlows",
    "StreamsetsConnection",
    "Environment",
    "Environments",
    "Engine",
]
